import numpy as np
import pandas as pd
import pytest

import spotify_confidence
from spotify_confidence.analysis.constants import METHOD_COLUMN_NAME, ZTEST, ADJUSTED_LOWER, ADJUSTED_UPPER


class TestBootstrap(object):
    def setup(self):
        np.random.seed(123)
        n_bootstraps = int(5e5)
        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test3"],
                "bootstraps": [
                    np.random.randn(n_bootstraps) + 0.5,
                    np.random.randn(n_bootstraps) + 0.4,
                    np.random.randn(n_bootstraps) + 0.2,
                    np.random.randn(n_bootstraps) + 1 / 3,
                ],
                "method": ["bootstrap", "bootstrap", "bootstrap", "bootstrap"],
            }
        )

        self.test = spotify_confidence.Experiment(
            self.data,
            numerator_column=None,
            numerator_sum_squares_column=None,
            denominator_column=None,
            categorical_group_columns="variation_name",
            method_column="method",
            bootstrap_samples_column="bootstraps",
        )

    def test_summary(self):
        summary_df = self.test.summary()
        assert len(summary_df) == len(self.data)

    def test_difference(self):
        difference_df = self.test.difference(level_1="control", level_2="test", absolute=True)
        assert len(difference_df) == 1
        assert np.allclose(difference_df["difference"][0], 0.1, atol=1e-2)

    def test_difference_absolute_false(self):
        difference_df = self.test.difference(level_1="control", level_2="test", absolute=False)
        assert len(difference_df) == 1
        assert np.allclose(difference_df["difference"][0], 0.1 / 0.4, atol=1e-2)

    def test_multiple_difference(self):
        difference_df = self.test.multiple_difference(level="control", level_as_reference=True)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"][0], 0.1, atol=1e-2)

    def test_multiple_difference_level_as_reference_false(self):
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=False, absolute=True)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"], -difference_df_true_true["difference"], atol=0)
        assert np.allclose(difference_df["ci_lower"], -difference_df_true_true["ci_upper"], atol=0)
        assert np.allclose(difference_df["ci_upper"], -difference_df_true_true["ci_lower"], atol=0)
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_multiple_difference_absolute_false(self):
        control_mean = self.test.summary().query("variation_name == 'control'")["point_estimate"].values[0]
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=True, absolute=False)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(difference_df["difference"], difference_df_true_true["difference"] / control_mean, atol=0)
        assert np.allclose(difference_df["ci_lower"], difference_df_true_true["ci_lower"] / control_mean, atol=0)
        assert np.allclose(difference_df["ci_upper"], difference_df_true_true["ci_upper"] / control_mean, atol=0)
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_multiple_difference_level_as_reference_false_absolute_false(self):
        reference_mean = self.test.summary().query("variation_name != 'control'")["point_estimate"]
        difference_df_true_true = self.test.multiple_difference(
            level="control", level_as_reference=True, absolute=True
        )

        difference_df = self.test.multiple_difference(level="control", level_as_reference=False, absolute=False)
        assert len(difference_df) == self.data.variation_name.unique().size - 1
        assert np.allclose(
            difference_df["difference"], -difference_df_true_true["difference"] / reference_mean.values, atol=0
        )
        assert np.allclose(
            difference_df["ci_lower"], -difference_df_true_true["ci_upper"] / reference_mean.values, atol=0
        )
        assert np.allclose(
            difference_df["ci_upper"], -difference_df_true_true["ci_lower"] / reference_mean.values, atol=0
        )
        assert np.allclose(difference_df["p-value"], difference_df_true_true["p-value"], atol=0)

    def test_summary_plot(self):
        chart_grid = self.test.summary_plot()
        assert len(chart_grid.charts) == 1

    def test_difference_plot(self):
        chartgrid = self.test.difference_plot(level_1="control", level_2="test")
        assert len(chartgrid.charts) == 1

    def test_multiple_difference_plot(self):
        chartgrid = self.test.multiple_difference_plot(level="control", level_as_reference=True)
        assert len(chartgrid.charts) == 1

    def test_differences_one_level_on_both_sides(self):
        df = pd.DataFrame(
            {
                "group_name": {0: "LT", 1: "HB", 2: "RNDM", 3: "ST"},
                "num_user": {0: 5832071, 1: 5830464, 2: 5775235, 3: 5829780},
                "sum": {
                    0: 3650267885.292686,
                    1: 3640976251.2644653,
                    2: 3543904424.4249864,
                    3: 3640408188.9692664,
                },
                "sum_squares": {
                    0: 11464986242442.066,
                    1: 11395508685623.664,
                    2: 10953117763878.217,
                    3: 11400833683366.701,
                },
            }
        )

        test = spotify_confidence.Experiment(
            data_frame=df.assign(**{METHOD_COLUMN_NAME: ZTEST}),
            numerator_column="sum",
            numerator_sum_squares_column="sum_squares",
            denominator_column="num_user",
            categorical_group_columns="group_name",
            method_column=METHOD_COLUMN_NAME,
        )

        diff = test.differences(
            levels=[
                ("RNDM", "HB"),
                ("ST", "HB"),
                ("LT", "HB"),
                ("ST", "LT"),
            ]
        )

        assert len(diff) == 4

    def test_multiple_differences_nims_false_none_true(self):
        pd.options.display.max_columns = None
        test = self.get_experiment_with_some_nims()

        diff_nim_false = test.multiple_difference(
            level="Control",
            level_as_reference=True,
            groupby="metric",
            non_inferiority_margins=False,
        )

        diff_nim_none = test.multiple_difference(
            level="Control",
            level_as_reference=True,
            groupby="metric",
            non_inferiority_margins=None,
        )

        assert np.allclose(diff_nim_false[ADJUSTED_LOWER], diff_nim_none[ADJUSTED_LOWER])
        assert np.allclose(diff_nim_false[ADJUSTED_UPPER], diff_nim_none[ADJUSTED_UPPER])

        diff_nim_true = test.multiple_difference(
            level="Control",
            level_as_reference=True,
            groupby="metric",
            non_inferiority_margins=True,
        )

        assert np.all(diff_nim_true[ADJUSTED_LOWER] > diff_nim_none[ADJUSTED_LOWER])
        assert np.all(diff_nim_true[ADJUSTED_UPPER] < diff_nim_none[ADJUSTED_UPPER])

    @pytest.mark.parametrize(
        "nims",
        [
            False,
            None,
            True,
        ],
        ids=lambda x: f"non_inferiority_margins: {x}",
    )
    def test_multiple_differences_plot_some_nims_doesnt_raise_exception(self, nims):
        pd.options.display.max_columns = None
        test = self.get_experiment_with_some_nims()
        try:
            ch = test.multiple_difference_plot(
                level="Control",
                level_as_reference=True,
                groupby="metric",
                absolute=True,
                use_adjusted_intervals=True,
                split_plot_by_groups=True,
                non_inferiority_margins=nims,
            )
            assert len(ch.charts) > 0
        except Exception as e:
            assert False, f"Using non_inferiority_margins={nims} raised an exception: {e}."

    def get_experiment_with_some_nims(self):
        columns = [
            "group_name",
            "num_user",
            "sum",
            "sum_squares",
            "method",
            "metric",
            "preferred_direction",
            "non_inferiority_margin",
        ]
        data = [
            ["Control", 6267728, 3240932, 3240932, "z-test", "m1", "increase", 0.15],
            ["Test", 6260737, 3239706, 3239706, "z-test", "m1", "increase", 0.15],
            ["Test", 6260737, 38600871, 12432573969, "z-test", "m2", "increase", None],
            ["Control", 6267728, 35963863, 18433512959, "z-test", "m2", "increase", None],
            ["Test", 6260737, 67382943, 8974188321, "z-test", "m3", "increase", None],
            ["Control", 6267728, 67111374, 8728934960, "z-test", "m3", "increase", None],
        ]
        df = pd.DataFrame(columns=columns, data=data)
        test = spotify_confidence.Experiment(
            data_frame=df.assign(**{METHOD_COLUMN_NAME: ZTEST}),
            numerator_column="sum",
            numerator_sum_squares_column="sum_squares",
            denominator_column="num_user",
            categorical_group_columns="metric",
            interval_size=0.99,
            correction_method="spot-1-bonferroni",
            metric_column="metric",
            treatment_column="group_name",
            method_column="method",
        )
        return test



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from functools import wraps
import types
import warnings

import chartify
import numpy as np
import pandas as pd

from spotify_confidence.options import options
from spotify_confidence.chartgrid import ChartGrid
from spotify_confidence.analysis.confidence_utils import de_list_if_length_one

# warnings.simplefilter("once")

INITIAL_RANDOMIZATION_SEED = np.random.get_state()[1][0]


def num_decimals(value: float, absolute: bool) -> int:
    extra_zeros = 3 if absolute else 1
    return -int(np.log10(abs(value))) + extra_zeros


def format_str_precision(value: float, absolute: bool) -> str:
    n = num_decimals(value, absolute)
    return f".{n}{'' if absolute else '%'}"


def axis_format_precision(max_value, min_value, absolute):
    zeros = "".join(["0"] * num_decimals(max_value - min_value, absolute))
    return "0.{}{}".format(zeros, "" if absolute else "%")


def add_color_column(df, cols):
    for i, column in enumerate(cols):
        if i == 0:
            df["color"] = df[column]
        else:
            df["color"] = df["color"] + " " + df[column]
    return df


def randomization_warning_decorator(f):
    """Set numpy randomization seed and warn users if not fixed.

    Note to developers:
    Do not compare random variables that have been
    sampled from the same seed. It will lead to incorrect results.
    To avoid this situation it's best to apply this decorator to
    public methods that involve randomization.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        option_seed = options.get_option("randomization_seed")
        np_seed = INITIAL_RANDOMIZATION_SEED
        if option_seed != np_seed and option_seed is None:
            randomization_warning_message = """
    Your analysis will not be reproducible!
    Using a method that involves randomization without setting a seed.
    Please run the following and add it to the top of your script or
    notebook after you import confidence:

    confidence.options.set_option('randomization_seed', {})

    """.format(
                INITIAL_RANDOMIZATION_SEED
            )
            warnings.warn(randomization_warning_message)
            option_seed = np_seed
        np.random.seed(option_seed)
        return f(*args, **kwargs)

    return wrapper


class BaseTest(object, metaclass=ABCMeta):
    """Base test class that provides abstract methods
    to ensure consistency across test classes."""

    def __init__(
        self,
        data_frame,
        categorical_group_columns,
        ordinal_group_column,
        numerator_column,
        denominator_column,
        interval_size,
    ):
        self._data_frame = data_frame
        self._numerator_column = numerator_column
        self._denominator_column = denominator_column
        self._interval_size = interval_size

        categorical_string_or_none = isinstance(categorical_group_columns, str) or categorical_group_columns is None
        self._categorical_group_columns = (
            [categorical_group_columns] if categorical_string_or_none else categorical_group_columns
        )
        self._ordinal_group_column = ordinal_group_column

        self._all_group_columns = self._categorical_group_columns + [self._ordinal_group_column]
        self._all_group_columns = [column for column in self._all_group_columns if column is not None]
        self._validate_data()

    def _validate_data(self):
        """Integrity check input dataframe."""
        if not self._all_group_columns:
            raise ValueError(
                """At least one of `categorical_group_columns`
                                or `ordinal_group_column` must be specified."""
            )

        # Ensure there's at most 1 observation per grouping.
        max_one_row_per_grouping = all(self._data_frame.groupby(self._all_group_columns).size() <= 1)
        if not max_one_row_per_grouping:
            raise ValueError("""Each grouping should have at most 1 observation.""")

        if self._ordinal_group_column:
            ordinal_column_type = self._data_frame[self._ordinal_group_column].dtype.type
            if not np.issubdtype(ordinal_column_type, np.number) and not issubclass(
                ordinal_column_type, np.datetime64
            ):
                raise TypeError(
                    """`ordinal_group_column` is type `{}`.
    Must be number or datetime type.""".format(
                        ordinal_column_type
                    )
                )

    @classmethod
    def as_cumulative(
        cls, data_frame, numerator_column, denominator_column, ordinal_group_column, categorical_group_columns=None
    ):
        """
        Instantiate the class with a cumulative representation of the dataframe.
        Sorts by the ordinal variable and calculates the cumulative sum
        May be used for to visualize the difference between groups as a
        time series.

        Args:
           data_frame (pd.DataFrame): DataFrame
           numerator_column (str): Column name for numerator column.
           denominator_column (str): Column name for denominator column.
           ordinal_group_column (str): Column name for ordinal grouping
               (e.g. numeric or date values).
           categorical_group_columns (str or list),
               Optional: Column names for categorical groupings.

        """

        sorted_df = data_frame.sort_values(by=ordinal_group_column)
        cumsum_cols = [numerator_column, denominator_column]
        if categorical_group_columns:
            sorted_df[cumsum_cols] = sorted_df.groupby(by=categorical_group_columns)[cumsum_cols].cumsum()
        else:
            sorted_df[cumsum_cols] = sorted_df[cumsum_cols].cumsum()

        return cls(sorted_df, numerator_column, denominator_column, categorical_group_columns, ordinal_group_column)

    def summary(self):
        """Return Pandas DataFrame with summary statistics."""
        return self._summary(self._data_frame, self._interval)

    def _summary(self, data_frame, ci_function):
        """Return the input dataframe with added columns:
        - Lower & upper bounds of
            Bayesian: credible interval
            Frequentist: confidence interval
        - Additional summary stats
            (e.g. probability in the case of Binomial data)
        """
        summary_df = data_frame[self._all_group_columns + [self._numerator_column, self._denominator_column]].copy()

        summary_df["point_estimate"] = summary_df[self._numerator_column] * 1.0 / summary_df[self._denominator_column]
        summary_df[["ci_lower", "ci_upper"]] = data_frame.apply(ci_function, axis=1, result_type="expand")

        return summary_df

    def summary_plot(self, groupby=None):
        """Plot for each group in the data_frame:

        if ordinal level exists:
            Frequentist: line graph with area to represent confidence interval
            Bayesian: line graph with area to represent credible interval
        if categorical levels:
            Bayesian: KDE plot of posterior distributions by group
            Frequentist: Interval plots of confidence intervals by group

        Args:
            groupby (str): Name of column.
                If specified, will plot a separate chart for each level of the
                grouping.

        Returns:
            ChartGrid object.
        """
        chart_grid = self._iterate_groupby_to_chartgrid(self._summary_plot, groupby=groupby)
        return chart_grid

    def _summary_plot(self, level_name, level_df, remaining_groups, groupby):
        if self._ordinal_group_column is not None and self._ordinal_group_column in remaining_groups:
            ch = self._ordinal_summary_plot(level_name, level_df, remaining_groups, groupby)
        else:
            ch = self._categorical_summary_plot(level_name, level_df, remaining_groups, groupby)
        return ch

    def _ordinal_summary_plot(self, level_name, level_df, remaining_groups, groupby):
        remaining_groups = self._remaining_categorical_groups(remaining_groups)
        df = self._summary(level_df, self._interval)
        title = "Estimate of {} / {}".format(self._numerator_column, self._denominator_column)
        y_axis_label = "{} / {}".format(self._numerator_column, self._denominator_column)
        return self._ordinal_plot(
            "point_estimate",
            df,
            groupby,
            level_name,
            remaining_groups,
            absolute=True,
            title=title,
            y_axis_label=y_axis_label,
        )

    def _ordinal_plot(self, center_name, df, groupby, level_name, remaining_groups, absolute, title, y_axis_label):
        df = add_color_column(df, remaining_groups)
        colors = "color" if remaining_groups else None
        ch = chartify.Chart(x_axis_type=self._ordinal_type())
        ch.plot.line(
            data_frame=df.sort_values(self._ordinal_group_column),
            x_column=self._ordinal_group_column,
            y_column=center_name,
            color_column=colors,
        )
        ch.style.color_palette.reset_palette_order()
        ch.plot.area(
            data_frame=df.sort_values(self._ordinal_group_column),
            x_column=self._ordinal_group_column,
            y_column="ci_lower",
            second_y_column="ci_upper",
            color_column=colors,
        )
        ch.axes.set_yaxis_label(y_axis_label)
        ch.axes.set_xaxis_label(self._ordinal_group_column)
        ch.set_source_label("")
        axis_format = axis_format_precision(df["ci_lower"].min(), df["ci_upper"].max(), absolute)
        ch.axes.set_yaxis_tick_format(axis_format)
        subtitle = "" if not groupby else "{}: {}".format(groupby, level_name)
        ch.set_subtitle(subtitle)
        ch.set_title(title)
        if colors:
            ch.set_legend_location("outside_bottom")
        return ch

    def _remaining_categorical_groups(self, remaining_groups):
        remaining_groups_list = [remaining_groups] if isinstance(remaining_groups, str) else remaining_groups

        remaining_categorical_groups = [
            group_name for group_name in remaining_groups_list if group_name != self._ordinal_group_column
        ]
        return remaining_categorical_groups

    def _ordinal_type(self):
        ordinal_column_type = self._data_frame[self._ordinal_group_column].dtype.type
        axis_type = "datetime" if issubclass(ordinal_column_type, np.datetime64) else "linear"
        return axis_type

    @abstractmethod
    def _categorical_summary_plot(self, level_name, level_df, remaining_groups, groupby):
        pass

    @abstractmethod
    def difference(self, level_1, level_2, absolute=True, groupby=None):
        """Return dataframe containing the difference in means between
            group 1 and 2 and the appropriate test statistics.
        Frequentist:
        - Calculate one of the following tests depending of the
            response variable type.
            - Binomial: Chisq / fisher exact test
            - Gaussian: t-test / z-test
            Return the p-value.
        Bayesian:
        - Calcuate the posterior distribution of the difference in means.
            Return the
            - probability that group 2 > group 1.
            - Expected loss
            - Expected change
            - Expected gain
            - 95% CI interval
        """
        pass

    def difference_plot(self, level_1, level_2, absolute=True, groupby=None):
        """Plot representing the difference between group 1 and 2.
        - Difference in means or proportions, depending
            on the response variable type.

        Frequentist:
        - Plot interval plot with confidence interval of the
            difference between groups

        Bayesian:
        - Plot KDE representing the posterior distribution of the difference.
            - Probability that group2 > group1
            - Mean difference
            - 95% interval.

        Args:
            level_1 (str, tuple of str): Name of first level.
            level_2 (str, tuple of str): Name of second level.
            absolute (bool): If True then return the absolute
                difference (level2 - level1)
                otherwise return the relative difference (level2 / level1 - 1)
            groupby (str): Name of column, or list of columns.
                If specified, will return an interval for each level
                of the grouped dimension, or a confidence band if the
                grouped dimension is ordinal

        Returns:
            GroupedChart object.
        """

        use_ordinal_axis = self._use_ordinal_axis(groupby)

        if use_ordinal_axis:
            ch = self._ordinal_difference_plot(level_1, level_2, absolute, groupby)
            chart_grid = ChartGrid()
            chart_grid.charts.append(ch)
        else:
            chart_grid = self._categorical_difference_plot(level_1, level_2, absolute, groupby)

        return chart_grid

    def _use_ordinal_axis(self, groupby):
        is_ordinal_difference_plot = (
            groupby is not None and self._ordinal_group_column is not None and self._ordinal_group_column in groupby
        )
        return is_ordinal_difference_plot

    def _ordinal_difference_plot(self, level_1, level_2, absolute, groupby):
        difference_df = self.difference(level_1, level_2, absolute, groupby)
        remaining_groups = self._remaining_categorical_groups(groupby)
        title = "Change from {} to {}".format(level_1, level_2)
        y_axis_label = self.get_difference_plot_label(absolute)
        ch = self._ordinal_plot(
            "difference",
            difference_df,
            groupby=None,
            level_name="",
            remaining_groups=remaining_groups,
            absolute=absolute,
            title=title,
            y_axis_label=y_axis_label,
        )
        ch.callout.line(0)

        return ch

    def get_difference_plot_label(self, absolute):
        change_type = "Absolute" if absolute else "Relative"
        return change_type + " change in {} / {}".format(self._numerator_column, self._denominator_column)

    @abstractmethod
    def _categorical_difference_plot(self, level_1, level_2, absolute, groupby):
        pass

    @abstractmethod
    def multiple_difference(self, level, absolute=True, groupby=None, level_as_reference=False):
        """The pairwise probability that the specific group
        is greater than all other groups.
        """
        pass

    def multiple_difference_plot(self, level, absolute=True, groupby=None, level_as_reference=False):
        """Compare level to all other groups or, if level_as_reference = True,
        all other groups to level.

        Args:
            level (str, tuple of str): Name of level.
            absolute (bool): If True then return the absolute
                difference (level2 - level1)
                otherwise return the relative difference (level2 / level1 - 1)
            groupby (str): Name of column, or list of columns.
                If specified, will return an interval for each level
                of the grouped dimension, or a confidence band if the
                grouped dimension is ordinal
            level_as_reference: If false, compare level to all other
             groups. If true, compare all other groups to level.
        """
        use_ordinal_axis = self._use_ordinal_axis(groupby)

        if use_ordinal_axis:
            ch = self._ordinal_multiple_difference_plot(level, absolute, groupby, level_as_reference)
            chart_grid = ChartGrid()
            chart_grid.charts.append(ch)
        else:
            chart_grid = self._categorical_multiple_difference_plot(level, absolute, groupby, level_as_reference)

        return chart_grid

    def _ordinal_multiple_difference_plot(self, level, absolute, groupby, level_as_reference):
        difference_df = self.multiple_difference(level, absolute, groupby, level_as_reference)
        remaining_groups = self._remaining_categorical_groups(groupby)
        groupby_columns = self._add_level_column(remaining_groups, level_as_reference)
        title = "Comparison to {}".format(level)
        y_axis_label = self.get_difference_plot_label(absolute)
        ch = self._ordinal_plot(
            "difference",
            difference_df,
            groupby=None,
            level_name="",
            remaining_groups=groupby_columns,
            absolute=absolute,
            title=title,
            y_axis_label=y_axis_label,
        )
        ch.callout.line(0)
        return ch

    def _add_level_column(self, groupby, level_as_reference):
        level_column = "level_2" if level_as_reference else "level_1"
        if groupby is None:
            groupby_columns = level_column
        else:
            if isinstance(groupby, str):
                groupby_columns = [groupby, level_column]
            else:
                groupby_columns = groupby + [level_column]
        return groupby_columns

    @abstractmethod
    def _categorical_multiple_difference_plot(self, level, absolute, groupby, level_as_reference):
        pass

    @staticmethod
    def _validate_levels(level_df, remaining_groups, level):
        try:
            level_df.groupby(remaining_groups).get_group(level)
        except (KeyError, ValueError):
            raise ValueError(
                """
                Invalid level: '{}'
                Must supply a level within the ungrouped dimensions: {}
                Valid levels:
                {}
                """.format(
                    level, remaining_groups, list(level_df.groupby(remaining_groups).groups.keys())
                )
            )

    def _groupby_iterator(self, input_function, groupby, **kwargs):
        groupby = [] if groupby is None else groupby
        # Will group over the whole dataframe if groupby is None
        level_groups = groupby if groupby else np.ones(len(self._data_frame))
        remaining_groups = [group for group in self._all_group_columns if group not in groupby and group is not None]

        for level_name, level_df in self._data_frame.groupby(de_list_if_length_one(level_groups)):
            yield input_function(level_name, level_df, remaining_groups, groupby, **kwargs)

    def _iterate_groupby_to_chartgrid(self, input_function, groupby, **kwargs):
        """Iterate through groups in the test and apply the input function.

        Returns ChartGrid"""
        chart_grid = ChartGrid()

        chart_grid.charts = list(self._groupby_iterator(input_function, groupby, **kwargs))

        return chart_grid

    def _iterate_groupby_to_dataframe(self, input_function, groupby, **kwargs):
        """Iterate through groups in the test and apply the input function.

        Returns pd.DataFrame"""
        groupby_iterator = self._groupby_iterator(input_function, groupby, **kwargs)

        # Flatten any nested generators.
        groupby_iterator = list(groupby_iterator)
        if isinstance(groupby_iterator[0], types.GeneratorType):
            groupby_iterator = [group for generator in groupby_iterator for group in generator]

        results_data_frame = pd.concat(groupby_iterator, axis=0)

        results_data_frame = results_data_frame.reset_index(drop=True)

        return results_data_frame

    def _all_groups(self):
        """Return a list of all group keys.

        Returns: list"""
        groups = list(self._data_frame.groupby(self._all_group_columns).groups.keys())
        return groups

    def _add_group_by_columns(self, difference_df, groupby, level_name):
        if groupby:
            groupby = groupby[0] if len(groupby) == 1 else groupby
            if isinstance(groupby, str):
                difference_df.insert(0, column=groupby, value=level_name)
            else:
                for col, val in zip(groupby, level_name):
                    difference_df.insert(0, column=col, value=val)


# class BinomialResponse(BaseTest, metaclass=ABCMeta):
#     """Binomial Response Variable.
#     """

# class GaussianResponse(BaseTest, metaclass=ABCMeta):
#     """Base class for tests of normal response variables

#     E.g. Revenue per user
#     """

#     pass


# class PoissonResponse(BaseTest, metaclass=ABCMeta):
#     """Base class for tests of poisson response variables.

#     E.g. # of days active per user per month
#     """
#     pass


# class MultinomialResponse(BaseTest, metaclass=ABCMeta):
#     """Base class for tests of multinomial response variables.

#     E.g. single choice answer survey
#         self.
#     """

#     def __init__(self, data_frame, categorical_group_columns,
#                  ordinal_group_column, category_column, value_column):
#         self._category_column = category_column
#         self._value_column = value_column
#         super().__init__(data_frame, categorical_group_columns,
#                          ordinal_group_column)


# class CategoricalResponse(BaseTest, metaclass=ABCMeta):
#     """Base class for tests of categorical response variables.

#     E.g. multiple choice answer survey
#     """

#     def __init__(self, data_frame, categorical_group_columns,
#                  ordinal_group_column, category_column, value_column):
#         self._category_column = category_column
#         self._value_column = value_column
#         super().__init__(data_frame, categorical_group_columns,
#                          ordinal_group_column)

#     pass



from _warnings import warn
from typing import Iterable, Dict

from pandas import DataFrame, Series
from statsmodels.stats.multitest import multipletests

from spotify_confidence.analysis.constants import (
    BONFERRONI,
    BONFERRONI_ONLY_COUNT_TWOSIDED,
    PREFERENCE_TEST,
    TWO_SIDED,
    HOLM,
    HOMMEL,
    SIMES_HOCHBERG,
    SIDAK,
    HOLM_SIDAK,
    FDR_BH,
    FDR_BY,
    FDR_TSBH,
    FDR_TSBKY,
    BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
    SPOT_1,
    SPOT_1_HOLM,
    SPOT_1_HOMMEL,
    SPOT_1_SIMES_HOCHBERG,
    SPOT_1_SIDAK,
    SPOT_1_HOLM_SIDAK,
    SPOT_1_FDR_BH,
    SPOT_1_FDR_BY,
    SPOT_1_FDR_TSBH,
    SPOT_1_FDR_TSBKY,
    NIM,
    NUMBER_OF_COMPARISONS,
    FINAL_EXPECTED_SAMPLE_SIZE,
    CORRECTION_METHOD,
    METHOD,
    IS_SIGNIFICANT,
    P_VALUE,
    ADJUSTED_ALPHA,
    ADJUSTED_P,
    ALPHA,
    INTERVAL_SIZE,
    ZTEST,
    ZTESTLINREG,
    CI_LOWER,
    CI_UPPER,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    PREFERENCE,
    ADJUSTED_ALPHA_POWER_SAMPLE_SIZE,
    CORRECTION_METHODS_THAT_REQUIRE_METRIC_INFO,
    ADJUSTED_POWER,
    POWER,
)
from spotify_confidence.analysis.frequentist.confidence_computers import confidence_computers


def get_num_comparisons(
    df: DataFrame,
    correction_method: str,
    number_of_level_comparisons: int,
    groupby: Iterable,
    metric_column: str,
    treatment_column: str,
    single_metric: bool,
    segments: Iterable,
) -> int:
    if correction_method == BONFERRONI:
        return max(
            1,
            number_of_level_comparisons * df.assign(_dummy_=1).groupby(groupby + ["_dummy_"], sort=False).ngroups,
        )
    elif correction_method == BONFERRONI_ONLY_COUNT_TWOSIDED:
        return max(
            number_of_level_comparisons
            * df.query(f'{PREFERENCE_TEST} == "{TWO_SIDED}"')
            .assign(_dummy_=1)
            .groupby(groupby + ["_dummy_"], sort=False)
            .ngroups,
            1,
        )
    elif correction_method in [
        HOLM,
        HOMMEL,
        SIMES_HOCHBERG,
        SIDAK,
        HOLM_SIDAK,
        FDR_BH,
        FDR_BY,
        FDR_TSBH,
        FDR_TSBKY,
    ]:
        return 1
    elif correction_method in [
        BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
        SPOT_1,
        SPOT_1_HOLM,
        SPOT_1_HOMMEL,
        SPOT_1_SIMES_HOCHBERG,
        SPOT_1_SIDAK,
        SPOT_1_HOLM_SIDAK,
        SPOT_1_FDR_BH,
        SPOT_1_FDR_BY,
        SPOT_1_FDR_TSBH,
        SPOT_1_FDR_TSBKY,
    ]:
        if metric_column is None or treatment_column is None:
            return max(
                1,
                number_of_level_comparisons
                * df[df[NIM].isnull()].assign(_dummy_=1).groupby(groupby + ["_dummy_"], sort=False).ngroups,
            )
        else:
            if single_metric:
                if df[df[NIM].isnull()].shape[0] > 0:
                    number_success_metrics = 1
                else:
                    number_success_metrics = 0
            else:
                number_success_metrics = df[df[NIM].isnull()].groupby(metric_column, sort=False).ngroups

            number_segments = (
                1
                if len(segments) == 0 or not all(item in df.index.names for item in segments)
                else df.groupby(segments, sort=False).ngroups
            )

            return max(1, number_of_level_comparisons * max(1, number_success_metrics) * number_segments)
    else:
        raise ValueError(f"Unsupported correction method: {correction_method}.")


def add_adjusted_p_and_is_significant(df: DataFrame, **kwargs: Dict) -> DataFrame:
    n_comparisons = kwargs[NUMBER_OF_COMPARISONS]
    if kwargs[FINAL_EXPECTED_SAMPLE_SIZE] is not None:
        if kwargs[CORRECTION_METHOD] not in [
            BONFERRONI,
            BONFERRONI_ONLY_COUNT_TWOSIDED,
            BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
            SPOT_1,
        ]:
            raise ValueError(
                f"{kwargs[CORRECTION_METHOD]} not supported for sequential tests. Use one of"
                f"{BONFERRONI}, {BONFERRONI_ONLY_COUNT_TWOSIDED}, "
                f"{BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY}, {SPOT_1}"
            )
        adjusted_alpha = compute_sequential_adjusted_alpha(df, **kwargs)
        df = df.merge(adjusted_alpha, left_index=True, right_index=True)
        df[IS_SIGNIFICANT] = df[P_VALUE] < df[ADJUSTED_ALPHA]
        df[P_VALUE] = None
        df[ADJUSTED_P] = None
    elif kwargs[CORRECTION_METHOD] in [
        HOLM,
        HOMMEL,
        SIMES_HOCHBERG,
        SIDAK,
        HOLM_SIDAK,
        FDR_BH,
        FDR_BY,
        FDR_TSBH,
        FDR_TSBKY,
        SPOT_1_HOLM,
        SPOT_1_HOMMEL,
        SPOT_1_SIMES_HOCHBERG,
        SPOT_1_SIDAK,
        SPOT_1_HOLM_SIDAK,
        SPOT_1_FDR_BH,
        SPOT_1_FDR_BY,
        SPOT_1_FDR_TSBH,
        SPOT_1_FDR_TSBKY,
    ]:
        if kwargs[CORRECTION_METHOD].startswith("spot-"):
            correction_method = kwargs[CORRECTION_METHOD][7:]
        else:
            correction_method = kwargs[CORRECTION_METHOD]
        df[ADJUSTED_ALPHA] = df[ALPHA] / n_comparisons
        is_significant, adjusted_p, _, _ = multipletests(
            pvals=df[P_VALUE], alpha=1 - kwargs[INTERVAL_SIZE], method=correction_method
        )
        df[ADJUSTED_P] = adjusted_p
        df[IS_SIGNIFICANT] = is_significant
    elif kwargs[CORRECTION_METHOD] in [
        BONFERRONI,
        BONFERRONI_ONLY_COUNT_TWOSIDED,
        BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
        SPOT_1,
    ]:
        df[ADJUSTED_ALPHA] = df[ALPHA] / n_comparisons
        df[ADJUSTED_P] = df[P_VALUE].map(lambda p: min(p * n_comparisons, 1))
        df[IS_SIGNIFICANT] = df[P_VALUE] < df[ADJUSTED_ALPHA]
    else:
        raise ValueError("Can't figure out which correction method to use :(")

    return df


def compute_sequential_adjusted_alpha(df: DataFrame, **kwargs: Dict) -> Series:
    if df[kwargs[METHOD]].isin([ZTEST, ZTESTLINREG]).all():
        return confidence_computers[ZTEST].compute_sequential_adjusted_alpha(df, **kwargs)
    else:
        raise NotImplementedError("Sequential testing is only supported for z-test and z-testlinreg")


def add_ci(df: DataFrame, **kwargs: Dict) -> DataFrame:
    lower, upper = confidence_computers[df[kwargs[METHOD]].values[0]].ci(df, ALPHA, **kwargs)

    if kwargs[CORRECTION_METHOD] in [
        HOLM,
        HOMMEL,
        SIMES_HOCHBERG,
        SPOT_1_HOLM,
        SPOT_1_HOMMEL,
        SPOT_1_SIMES_HOCHBERG,
    ] and all(df[PREFERENCE_TEST] != TWO_SIDED):
        if all(df[kwargs[METHOD]] == "z-test"):
            adjusted_lower, adjusted_upper = confidence_computers["z-test"].ci_for_multiple_comparison_methods(
                df, kwargs[CORRECTION_METHOD], alpha=1 - kwargs[INTERVAL_SIZE]
            )
        else:
            raise NotImplementedError(f"{kwargs[CORRECTION_METHOD]} is only supported for ZTests")
    elif kwargs[CORRECTION_METHOD] in [
        BONFERRONI,
        BONFERRONI_ONLY_COUNT_TWOSIDED,
        BONFERRONI_DO_NOT_COUNT_NON_INFERIORITY,
        SPOT_1,
        SPOT_1_HOLM,
        SPOT_1_HOMMEL,
        SPOT_1_SIMES_HOCHBERG,
        SPOT_1_SIDAK,
        SPOT_1_HOLM_SIDAK,
        SPOT_1_FDR_BH,
        SPOT_1_FDR_BY,
        SPOT_1_FDR_TSBH,
        SPOT_1_FDR_TSBKY,
    ]:
        adjusted_lower, adjusted_upper = confidence_computers[df[kwargs[METHOD]].values[0]].ci(
            df, ADJUSTED_ALPHA, **kwargs
        )
    else:
        warn(f"Confidence intervals not supported for {kwargs[CORRECTION_METHOD]}")
        adjusted_lower = None
        adjusted_upper = None

    return (
        df.assign(**{CI_LOWER: lower})
        .assign(**{CI_UPPER: upper})
        .assign(**{ADJUSTED_LOWER: adjusted_lower})
        .assign(**{ADJUSTED_UPPER: adjusted_upper})
    )


def set_alpha_and_adjust_preference(df: DataFrame, **kwargs: Dict) -> DataFrame:
    alpha_0 = 1 - kwargs[INTERVAL_SIZE]
    return df.assign(
        **{
            ALPHA: df.apply(
                lambda row: 2 * alpha_0
                if kwargs[CORRECTION_METHOD] == SPOT_1 and row[PREFERENCE] != TWO_SIDED
                else alpha_0,
                axis=1,
            )
        }
    ).assign(**{ADJUSTED_ALPHA_POWER_SAMPLE_SIZE: lambda df: df[ALPHA] / kwargs[NUMBER_OF_COMPARISONS]})


def get_preference(df: DataFrame, correction_method: str):
    return TWO_SIDED if correction_method == SPOT_1 else df[PREFERENCE]


def add_adjusted_power(df: DataFrame, correction_method: str, metric_column: str, single_metric: bool) -> DataFrame:
    if correction_method in CORRECTION_METHODS_THAT_REQUIRE_METRIC_INFO:
        if metric_column is None:
            return df.assign(**{ADJUSTED_POWER: None})
        else:
            number_total_metrics = 1 if single_metric else df.groupby(metric_column, sort=False).ngroups
            if single_metric:
                if df[df[NIM].isnull()].shape[0] > 0:
                    number_success_metrics = 1
                else:
                    number_success_metrics = 0
            else:
                number_success_metrics = df[df[NIM].isnull()].groupby(metric_column, sort=False).ngroups

            number_guardrail_metrics = number_total_metrics - number_success_metrics
            power_correction = (
                number_guardrail_metrics if number_success_metrics == 0 else number_guardrail_metrics + 1
            )
            return df.assign(**{ADJUSTED_POWER: 1 - (1 - df[POWER]) / power_correction})
    else:
        return df.assign(**{ADJUSTED_POWER: df[POWER]})



"""Tests for `confidence` categorical variables."""

import numpy as np
import pandas as pd
import pytest
from scipy.stats import chisquare

import spotify_confidence
from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import POINT_ESTIMATE, VARIANCE, SFX1, SFX2, DENOMINATOR
from spotify_confidence.analysis.frequentist.confidence_computers import chi_squared_computer as computer


def chart_data(chart_object, series_name):
    """Retrieve data from chart object based on series name.

    Note: if there's only one series the name is None.

    Returns:
        Dictionary populated with data from the chart.
    """
    cannonical_series_name = chart_object.plot._cannonical_series_name(series_name)
    return chart_object.figure.select(cannonical_series_name)[0].data


class TestCategorical(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "success": [500, 42, 1005, 50, 40, 100],
                "total": [1009, 104, 1502, 100, 100, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.ChiSquared(
            self.data,
            numerator_column="success",
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_bernoulli_squared_standard_error(self):
        n = 1000
        sample_size = int(1e7)
        p = 0.2

        x = np.random.binomial(n, p, sample_size) / n
        se = x.var()

        se_to_verify = pd.DataFrame({POINT_ESTIMATE: [p]}).apply(computer.variance, axis=1, arg_dict=None) / n

        assert np.allclose(se_to_verify, se, atol=1e-6)

    def test_standard_error_diff_prop(self):
        sample_size = int(1e7)

        p1 = 0.1
        n1 = int(1e6)
        x1 = np.random.binomial(n1, p1, sample_size) / n1

        n2 = int(2e6)
        p2 = 0.2
        x2 = np.random.binomial(n2, p2, sample_size) / n2

        diff = x2 - x1
        std_diff = diff.std()

        diff_se = computer.std_err(
            pd.DataFrame(
                {
                    VARIANCE + SFX1: [p1 * (1 - p1)],
                    VARIANCE + SFX2: [p2 * (1 - p2)],
                    "n" + SFX1: [n1],
                    "n" + SFX2: [n2],
                    "method": ["chi-squared"],
                }
            ),
            **{DENOMINATOR: "n"},
        )

        assert np.allclose(std_diff, diff_se, atol=1e-6)

    def test_summary(self):
        """Area plot tests"""

        summary = self.test.summary()
        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.success / self.data.total)
        assert np.allclose(
            summary["ci_lower"],
            np.array(
                [0.4646901340180582, 0.309544, 0.6453118311511006, 0.4020018007729973, 0.303982, 0.5912276177282552]
            ),
        )
        assert np.allclose(
            summary["ci_upper"],
            np.array(
                [0.5263901434844195, 0.498147, 0.692903881232388, 0.5979981992270027, 0.496018, 0.7421057156050781]
            ),
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1
        ch = self.test.summary_plot(groupby="country")
        assert len(ch.charts) == 2

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"))

        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], 0.091694)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.0075254411815593725)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.190913)
        assert np.allclose(diff["p-value"].iloc[0], 0.074866)

        diff = self.test.difference(("us", "control"), ("us", "test"), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], 0.227052)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.0186344)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.47273797)
        assert np.allclose(diff["p-value"].iloc[0], 0.074866)

        diff = self.test.difference("control", "test", groupby="country")
        assert np.allclose(diff["difference"], np.array([0.100000, 0.091694]))
        assert np.allclose(diff["ci_lower"], np.array([-0.03719748, -0.0075254]))
        assert np.allclose(diff["ci_upper"], np.array([0.237197, 0.190913]))
        assert np.allclose(diff["p-value"], np.array([0.155218, 0.074866]))

    def test_difference_with_interval_sizes(self):
        self.test._confidence_computer._interval_size = 0.99
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], 0.091694)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.0387024)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.22209)
        assert np.allclose(diff["p-value"].iloc[0], 0.074866)

        self.test._confidence_computer._interval_size = 0.999
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], 0.091694)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.074883)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.258271)
        assert np.allclose(diff["p-value"].iloc[0], 0.074866)

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("bad_value", "bad_value"), ("bad_value", "bad_value"))

        ch = self.test.difference_plot(("us", "control"), ("us", "test"))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="country")
        assert len(ch.charts) == 1

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(("bad_value", "bad_value"), level_as_reference=False)

        diff = self.test.multiple_difference(("us", "control"), level_as_reference=False)
        assert np.allclose(
            diff["adjusted p-value"],
            np.array([1.00000000e00, 8.38342202e-01, 1.67371350e-04, 3.74330946e-01, 1.98857657e-07]),
        )
        assert np.allclose(
            diff["p-value"], np.array([9.553332e-01, 1.67668440e-01, 3.34742700e-05, 7.48661891e-02, 3.97715314e-08])
        )
        assert np.allclose(
            diff["adjusted ci_lower"], np.array([-0.173024, -0.27489017, -0.42153065, -0.22209041, -0.39307973])
        )
        assert np.allclose(
            diff["adjusted ci_upper"], np.array([0.180717, 0.08258247, -0.10411038, 0.03870244, -0.13744367])
        )

        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=False)
        assert np.allclose(
            diff["adjusted p-value"], np.array([6.208740e-01, 3.36319783e-02, 2.99464756e-01, 1.30744685e-17])
        )
        assert np.allclose(diff["p-value"], np.array([1.552185e-01, 8.40799458e-03, 7.48661891e-02, 3.26861713e-18]))
        assert np.allclose(diff["adjusted ci_lower"], np.array([-0.074839, -0.32426934, -0.03474758, -0.2232184]))
        assert np.allclose(diff["adjusted ci_upper"], np.array([0.274839, -0.00906399, 0.21813554, -0.12391703]))

    def test_differences(self):
        with pytest.raises(ValueError):
            self.test.differences(("bad_value", "bad_value"))

        diff = self.test.differences(
            [
                (("gb", "control"), ("us", "control")),
                (("gb", "test"), ("us", "control")),
                (("gb", "test2"), ("us", "control")),
                (("us", "test"), ("us", "control")),
                (("us", "test2"), ("us", "control")),
            ]
        )
        assert np.allclose(
            diff["adjusted p-value"],
            np.array([1.00000000e00, 8.38342202e-01, 1.67371350e-04, 3.74330946e-01, 1.98857657e-07]),
        )
        assert np.allclose(
            diff["p-value"], np.array([9.553332e-01, 1.67668440e-01, 3.34742700e-05, 7.48661891e-02, 3.97715314e-08])
        )
        assert np.allclose(
            diff["adjusted ci_lower"], np.array([-0.173024, -0.27489017, -0.42153065, -0.22209041, -0.39307973])
        )
        assert np.allclose(
            diff["adjusted ci_upper"], np.array([0.180717, 0.08258247, -0.10411038, 0.03870244, -0.13744367])
        )

        diff = self.test.differences([("control", "test"), ("test2", "test")], groupby="country")
        assert np.allclose(
            diff["adjusted p-value"], np.array([6.208740e-01, 3.36319783e-02, 2.99464756e-01, 1.30744685e-17])
        )
        assert np.allclose(diff["p-value"], np.array([1.552185e-01, 8.40799458e-03, 7.48661891e-02, 3.26861713e-18]))
        assert np.allclose(diff["adjusted ci_lower"], np.array([-0.074839, -0.32426934, -0.03474758, -0.2232184]))
        assert np.allclose(diff["adjusted ci_upper"], np.array([0.274839, -0.00906399, 0.21813554, -0.12391703]))

    def test_multiple_difference_level_as_reference(self):
        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=True)
        assert np.allclose(diff["adjusted ci_lower"], np.array([-0.274839, 0.00906399, -0.21813554, 0.12391703]))

    def test_differences_plot(self):
        ch = self.test.differences_plot(
            [
                (("gb", "control"), ("us", "control")),
                (("gb", "test"), ("us", "control")),
                (("gb", "test2"), ("us", "control")),
                (("us", "test"), ("us", "control")),
                (("us", "test2"), ("us", "control")),
            ]
        )
        assert len(ch.charts) == 1

        ch = self.test.differences_plot([("control", "test"), ("test2", "test")], groupby="country")
        assert len(ch.charts) == 1

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        ch = self.test.multiple_difference_plot(("us", "control"), level_as_reference=False)
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot("test", groupby="country", level_as_reference=False)
        assert len(ch.charts) == 1

    def test_sample_ratio_test(self):
        expected = {
            ("us", "test"): 0.35,
            ("us", "control"): 0.04,
            ("us", "test2"): 0.52,
            ("gb", "test"): 0.035,
            ("gb", "control"): 0.0035,
            ("gb", "test2"): 0.0515,
        }

        with pytest.raises(TypeError):
            self.test.sample_ratio_test("not a dict")

        with pytest.raises(ValueError):
            negative = expected.copy()
            negative[("us", "test2")] = -0.1
            self.test.sample_ratio_test(negative)

        with pytest.raises(ValueError):
            not_one = expected.copy()
            not_one[("us", "test2")] += 0.1
            self.test.sample_ratio_test(not_one)

        with pytest.raises(ValueError):
            bad_group = expected.copy()
            bad_group[("X", "test2")] = bad_group.pop(("us", "test2"))
            self.test.sample_ratio_test(bad_group)

        p, df = self.test.sample_ratio_test(expected)

        obs_actual = df[self.test._denominator]
        obs_expected = sum(obs_actual) * df.expected_proportion
        _, p_ref = chisquare(obs_actual, obs_expected)

        assert np.allclose(p, p_ref)

    def test_power_calculation_binomial(self):
        mde = 0.01
        baseline_var = 0.34 * (1 - 0.34)
        alpha = 0.03
        n1 = 225725
        n2 = 22572
        power = power_calculation(mde, baseline_var, alpha, n1, n2)

        expected_power = 0.8  # Based on G*Power calculation for 80% power

        assert np.allclose(power, expected_power, 0.005)

    def test_power_calculation_continuous(self):
        mde = 0.03
        baseline_var = 45
        alpha = 0.05
        n1 = 859896
        n2 = 859896
        power = power_calculation(mde, baseline_var, alpha, n1, n2)

        expected_power = 0.8345828  # Based on G*Power calculation

        assert np.allclose(power, expected_power, 0.005)

    def test_achieved_power(self):
        power_df = self.test.achieved_power(
            level_1=("us", "control"), level_2=("us", "test"), mde=0.1429118527, alpha=0.05
        )
        power = power_df.achieved_power
        expected_power = 0.8  # Based on G*Power calculation
        assert np.allclose(power, expected_power, 0.01)

    def test_achieved_power_groupby(self):
        power_df = self.test.achieved_power(
            level_1="control", level_2="test", mde=0.1429118527, alpha=0.05, groupby="country"
        )
        power = power_df.achieved_power
        expected_power = [0.53, 0.800]  # Based on G*Power calculation
        assert np.allclose(power, expected_power, 0.01)


class TestOrdinal(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "success": [500, 8, 100, 510, 8, 100, 520, 9, 104, 530, 7, 100, 530, 8, 103],
                "total": [1010, 22, 150, 1000, 20, 153, 1030, 23, 154, 1000, 20, 150, 1040, 21, 155],
                "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
            }
        )

        self.test = spotify_confidence.ChiSquared(
            self.data,
            numerator_column="success",
            denominator_column="total",
            categorical_group_columns="variation_name",
            ordinal_group_column="days_since_reg",
        )

    def test_summary(self):
        """Area plot tests"""
        summary = self.test.summary()
        assert np.array_equal(summary["point_estimate"], self.data.success / self.data.total)
        assert np.allclose(
            summary["ci_lower"],
            np.array(
                [
                    0.46421506,
                    0.16262386,
                    0.59122762,
                    0.47901645,
                    0.1852967,
                    0.57819867,
                    0.4743207,
                    0.19185087,
                    0.60136948,
                    0.49906608,
                    0.1409627,
                    0.59122762,
                    0.47923305,
                    0.17325271,
                    0.59018498,
                ]
            ),
        )
        assert np.allclose(
            summary["ci_upper"],
            np.array(
                [
                    0.52588395,
                    0.56464887,
                    0.74210572,
                    0.54098355,
                    0.6147033,
                    0.72899087,
                    0.53538804,
                    0.59075782,
                    0.74927987,
                    0.56093392,
                    0.5590373,
                    0.74210572,
                    0.53999772,
                    0.58865206,
                    0.73884727,
                ]
            ),
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot().charts[0]
        np.array_equal(chart_data(ch, "control")["days_since_reg"], np.array([1, 2, 3, 4, 5, 5, 4, 3, 2, 1]))

    def test_p_value_is_symmetric(self):
        diff1 = self.test.difference(("control", 1), ("test", 1))
        assert np.allclose(diff1["difference"].iloc[0], 0.131413)
        assert np.allclose(diff1["p-value"].iloc[0], 0.222578)

        diff2 = self.test.difference(("test", 1), ("control", 1))
        assert np.allclose(diff2["difference"].iloc[0], -0.131413)
        assert np.allclose(diff2["p-value"].iloc[0], 0.222578)

        assert diff1["p-value"].iloc[0] == diff2["p-value"].iloc[0]

    def test_raise_error_with_nim(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", 1), ("test", 1), non_inferiority_margins=("blah", "hah"))

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"))

        diff = self.test.difference(("control", 1), ("test", 1))
        assert np.allclose(diff["difference"].iloc[0], 0.131413)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.071951)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.334777)
        assert np.allclose(diff["p-value"].iloc[0], 0.222578)

        diff = self.test.difference(("control", 1), ("test", 1), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], 0.361386)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.1978640)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.92063629)
        assert np.allclose(diff["p-value"].iloc[0], 0.222578)

        diff = self.test.difference("control", "test", groupby="days_since_reg")
        assert np.allclose(diff["difference"], np.array([0.13141314, 0.11, 0.11355002, 0.18, 0.128663]))

        diff = self.test.difference(("control", 1), ("test", 1))
        assert np.allclose(diff["difference"].iloc[0], 0.131413)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.071951)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.334777)
        assert np.allclose(diff["p-value"].iloc[0], 0.222578)

        diff = self.test.difference(("control", 1), ("test", 1), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], 0.361386)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.197864)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.9206363)
        assert np.allclose(diff["p-value"].iloc[0], 0.222578)

        diff = self.test.difference("control", "test", groupby="days_since_reg")
        assert np.allclose(diff["difference"], np.array([0.13141314, 0.11, 0.11355002, 0.18, 0.128663]))

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", "us"), ("test", "usf"))

        ch = self.test.difference_plot(("control", 1), ("test", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="days_since_reg")
        assert len(ch.charts) == 1

    def test_sample_ratio_test(self):
        expected = {
            ("control", 1): 0.004,
            ("control", 2): 0.003,
            ("control", 3): 0.004,
            ("control", 4): 0.003,
            ("control", 5): 0.004,
            ("test", 1): 0.15,
            ("test", 2): 0.168,
            ("test", 3): 0.173,
            ("test", 4): 0.168,
            ("test", 5): 0.195,
            ("test2", 1): 0.025,
            ("test2", 2): 0.026,
            ("test2", 3): 0.026,
            ("test2", 4): 0.025,
            ("test2", 5): 0.026,
        }

        p, df = self.test.sample_ratio_test(expected)

        obs_actual = df[self.test._denominator]
        obs_expected = sum(obs_actual) * df.expected_proportion
        _, p_ref = chisquare(obs_actual, obs_expected)

        assert np.allclose(p, p_ref)


class TestOrdinalPlusTwoCategorical(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "success": [
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                ],
                "total": [
                    2010,
                    42,
                    250,
                    2000,
                    40,
                    253,
                    2030,
                    43,
                    254,
                    2000,
                    40,
                    250,
                    2040,
                    41,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "days_since_reg": [
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                ],
                "country": [
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.ChiSquared(
            self.data,
            numerator_column="success",
            denominator_column="total",
            categorical_group_columns=["variation_name", "country"],
            ordinal_group_column="days_since_reg",
        )

    def test_summary_plot(self):
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot(groupby=["country"])
        assert len(ch.charts) == 2

        ch = self.test.summary_plot(groupby=["days_since_reg"])
        assert len(ch.charts) == 5

        ch = self.test.summary_plot(groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 10

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"), groupby="days_since_reg")

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby="days_since_reg")
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby=["days_since_reg"])
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", 1), level_2=("control", 1), groupby=["country"])
        assert len(df) == 2
        assert "country" in df.columns

        df = self.test.difference(level_1="test", level_2="control", groupby=["country", "days_since_reg"])
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(level="non_existingg_level", level_as_reference=False)

        df = self.test.multiple_difference(
            level="control", groupby=["country", "days_since_reg"], level_as_reference=False
        )
        assert len(df) == 20

        df = self.test.multiple_difference(
            level="us", groupby=["variation_name", "days_since_reg"], level_as_reference=False
        )
        assert len(df) == 15

        df = self.test.multiple_difference(level=1, groupby=["country", "variation_name"], level_as_reference=False)
        assert len(df) == 24

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", "us", 10), ("test", "us", 10))

        ch = self.test.difference_plot(level_1=("control", "us", 1), level_2=("test", "us", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", 1), level_2=("test", 1), groupby="country")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(level="control")

        ch = self.test.multiple_difference_plot(
            level="control", groupby=["country", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="control", groupby=["country", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level=1, groupby=["country", "variation_name"], level_as_reference=True
        )
        assert len(ch.charts) == 1



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Iterable

from pandas import DataFrame

from spotify_confidence.analysis.abstract_base_classes.confidence_computer_abc import ConfidenceComputerABC
from spotify_confidence.analysis.abstract_base_classes.confidence_grapher_abc import ConfidenceGrapherABC
from spotify_confidence.analysis.constants import BONFERRONI, METHOD_COLUMN_NAME
from spotify_confidence.analysis.frequentist.experiment import Experiment


class ChiSquared(Experiment):
    def __init__(
        self,
        data_frame: DataFrame,
        numerator_column: str,
        denominator_column: str,
        categorical_group_columns: Union[str, Iterable],
        ordinal_group_column: Union[str, None] = None,
        interval_size: float = 0.95,
        correction_method: str = BONFERRONI,
        confidence_computer: ConfidenceComputerABC = None,
        confidence_grapher: ConfidenceGrapherABC = None,
        metric_column: Union[str, None] = None,
        treatment_column: Union[str, None] = None,
    ):
        super(ChiSquared, self).__init__(
            data_frame=data_frame.assign(**{METHOD_COLUMN_NAME: "chi-squared"}),
            numerator_column=numerator_column,
            numerator_sum_squares_column=numerator_column,
            denominator_column=denominator_column,
            categorical_group_columns=categorical_group_columns,
            ordinal_group_column=ordinal_group_column,
            interval_size=interval_size,
            correction_method=correction_method,
            confidence_computer=confidence_computer,
            confidence_grapher=confidence_grapher,
            method_column=METHOD_COLUMN_NAME,
            metric_column=metric_column,
            treatment_column=treatment_column,
            power=0.8,
        )



from typing import Tuple, Dict

import numpy as np
from pandas import DataFrame, Series

from spotify_confidence.analysis.constants import CI_LOWER, CI_UPPER, SFX1, SFX2, BOOTSTRAPS, INTERVAL_SIZE


def point_estimate(df: DataFrame, **kwargs: Dict[str, str]) -> float:
    bootstrap_samples = kwargs[BOOTSTRAPS]
    return df[bootstrap_samples].map(lambda a: a.mean())


def variance(df: Series, **kwargs: Dict[str, str]) -> float:
    bootstrap_samples = kwargs[BOOTSTRAPS]
    variance = df[bootstrap_samples].map(lambda a: a.var())

    if (variance < 0).any():
        raise ValueError("Computed variance is negative. " "Please check your inputs.")
    return variance


def std_err(row: Series, **kwargs: Dict[str, str]) -> float:
    return None


def add_point_estimate_ci(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    bootstrap_samples = kwargs[BOOTSTRAPS]
    interval_size = kwargs[INTERVAL_SIZE]
    df[CI_LOWER] = df[bootstrap_samples].map(lambda a: np.percentile(a, 100 * (1 - interval_size) / 2))
    df[CI_UPPER] = df[bootstrap_samples].map(lambda a: np.percentile(a, 100 * (1 - (1 - interval_size) / 2)))
    return df


def p_value(row, **kwargs: Dict[str, str]) -> float:
    return -1


def ci(df, alpha_column: str, **kwargs: Dict[str, str]) -> Tuple[Series, Series]:
    bootstrap_samples = kwargs[BOOTSTRAPS]
    lower = df.apply(
        lambda row: np.percentile(
            row[bootstrap_samples + SFX2] - row[bootstrap_samples + SFX1], 100 * row[alpha_column] / 2
        ),
        axis=1,
    )
    upper = df.apply(
        lambda row: np.percentile(
            row[bootstrap_samples + SFX2] - row[bootstrap_samples + SFX1], 100 * (1 - row[alpha_column] / 2)
        ),
        axis=1,
    )
    return lower, upper


def achieved_power(df: DataFrame, mde: float, alpha: float) -> DataFrame:
    return None



"""Tests for `confidence` sample size calculation."""

import pytest
import spotify_confidence as conf
import numpy as np


class TestSampleSizeCalc(object):
    def test_continuous_default(self):
        n_tot, _, allocations = conf.SampleSize.continuous(average_absolute_mde=0.3, baseline_variance=100)
        g_power_n_tot = 39906
        #: Reference sample size calculated with G*Power.
        assert np.allclose(n_tot, g_power_n_tot, 0.005) and (allocations == [0.5, 0.5]).all()

    def test_binomial_default_small(self):
        n_tot, _, allocations = conf.SampleSize.binomial(absolute_percentage_mde=0.003, baseline_proportion=0.34)

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 897348
        assert np.allclose(n_tot, g_power_n_tot, 0.005) and (allocations == [0.5, 0.5]).all()

    def test_binomial_default_large(self):
        n_tot, _, allocations = conf.SampleSize.binomial(absolute_percentage_mde=0.1, baseline_proportion=0.34)

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = np.average((738, 852))
        assert np.allclose(n_tot, g_power_n_tot, 0.015) and (allocations == [0.5, 0.5]).all()

    def test_binomial_alpha(self):
        n_tot, _, _ = conf.SampleSize.binomial(absolute_percentage_mde=0.003, baseline_proportion=0.34, alpha=0.14)

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 630710
        assert np.allclose(n_tot, g_power_n_tot, 0.005)

    def test_binomial_alpha_power(self):
        n_tot, _, _ = conf.SampleSize.binomial(
            absolute_percentage_mde=0.003, baseline_proportion=0.34, alpha=0.14, power=0.67
        )

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 366424
        assert np.allclose(n_tot, g_power_n_tot, 0.005)

    def test_binomial_treatments(self):
        n_tot, _, allocations = conf.SampleSize.binomial(
            absolute_percentage_mde=0.003, baseline_proportion=0.34, treatments=3
        )

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 673157 + 336578 * 2
        assert np.allclose(n_tot, g_power_n_tot, 0.005) and (allocations == [1 / 2, 1 / 4, 1 / 4]).all()

    def test_binomial_treatments_all_vs_all(self):
        n_tot, _, allocations = conf.SampleSize.binomial(
            absolute_percentage_mde=0.003, baseline_proportion=0.34, treatments=3, comparisons="all_vs_all"
        )

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 448674 * 3
        assert np.allclose(n_tot, g_power_n_tot, 0.005) and (allocations == [1 / 3, 1 / 3, 1 / 3]).all()

    def test_continuous_treatments(self):
        n_tot, _, allocations = conf.SampleSize.continuous(
            average_absolute_mde=0.3, baseline_variance=100, treatments=3
        )

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 29929 + 14965 * 2
        assert np.allclose(n_tot, g_power_n_tot, 0.005) and (allocations == [1 / 2, 1 / 4, 1 / 4]).all()

    def test_continuous_treatments_all_vs_all(self):
        n_tot, _, allocations = conf.SampleSize.continuous(
            average_absolute_mde=0.3, baseline_variance=100, treatments=3, comparisons="all_vs_all"
        )

        #: Reference sample size calculated with G*Power.
        g_power_n_tot = 19953 * 3
        assert np.allclose(n_tot, g_power_n_tot, 0.005) and (allocations == [1 / 3, 1 / 3, 1 / 3]).all()

    def test_clean_treatments_int(self):
        returned_treatments = conf.SampleSize._clean_treatments(4)
        assert returned_treatments == 4

    def test_clean_treatments_negative(self):
        with pytest.raises(ValueError):
            conf.SampleSize._clean_treatments(-4)

    def test_clean_treatments_non_integer(self):
        with pytest.raises(ValueError):
            conf.SampleSize._clean_treatments(4.1)

    def test_clean_treatments_string(self):
        with pytest.raises(TypeError):
            conf.SampleSize._clean_treatments("X")

    def test_clean_comparisons_control_vs_all(self):
        comparison = "control_vs_all"
        returned_comparisons = conf.SampleSize._clean_comparisons(comparison)
        assert comparison == returned_comparisons

    def test_clean_comparisons_all_vs_all(self):
        comparison = "all_vs_all"
        returned_comparisons = conf.SampleSize._clean_comparisons(comparison)
        assert comparison == returned_comparisons

    def test_clean_comparisons_wrong_input(self):
        with pytest.raises(ValueError):
            conf.SampleSize._clean_comparisons("X")

    def test_num_comparisons_control_vs_all(self):
        comparisons = conf.SampleSize._num_comparisons(5, "control_vs_all")
        assert comparisons == 4

    def test_num_comparisons_all_vs_all_two(self):
        comparisons = conf.SampleSize._num_comparisons(2, "all_vs_all")
        assert comparisons == 1

    def test_num_comparisons_all_vs_all_five(self):
        comparisons = conf.SampleSize._num_comparisons(5, "all_vs_all")
        assert comparisons == 10

    def test_get_comparison_matrix_control_vs_all(self):
        returned_comparisons = conf.SampleSize._get_comparison_matrix(4, "control_vs_all")
        expected_comparisons = np.array(
            [[0.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0]]
        )
        assert np.array_equal(returned_comparisons, expected_comparisons)

    def test_get_comparison_matrix_all_vs_all(self):
        returned_comparisons = conf.SampleSize._get_comparison_matrix(4, "all_vs_all")
        expected_comparisons = np.array(
            [[0.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0], [1.0, 1.0, 0.0, 0.0], [1.0, 1.0, 1.0, 0.0]]
        )
        assert np.array_equal(returned_comparisons, expected_comparisons)

    def test_clean_treatment_costs_default(self):
        cost_array = conf.SampleSize._clean_treatment_costs(3, None)
        assert np.array_equal(np.ones(3), cost_array)

    def test_clean_treatment_costs_custom(self):
        custom_treatment_costs = np.array((1, 2, 3))
        returned_treatment_costs = conf.SampleSize._clean_treatment_costs(3, custom_treatment_costs)
        assert np.array_equal(custom_treatment_costs, returned_treatment_costs)

    def test_clean_treatment_costs_negative(self):
        with pytest.raises(ValueError):
            cost_array = np.array([0.3, 1, -2])
            conf.SampleSize._clean_treatment_costs(3, cost_array)

    def test_clean_treatment_costs_contains_string(self):
        with pytest.raises(TypeError):
            cost_array = np.array([0.3, 1, "X"])
            conf.SampleSize._clean_treatment_costs(3, cost_array)

    def test_clean_treatment_costs_shape(self):
        with pytest.raises(TypeError):
            cost_array = np.array([0.3, 1, 2])
            conf.SampleSize._clean_treatment_costs(4, cost_array)

    def test_clean_treatment_costs_int(self):
        with pytest.raises(TypeError):
            cost_array = 2
            conf.SampleSize._clean_treatment_costs(4, cost_array)

    def test_get_treatment_allocations_auto_control_vs_all(self):
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(3, "control_vs_all", None, None)
        expected_treatment_allocations = np.array((1 / 2, 1 / 4, 1 / 4))
        assert np.array_equal(returned_treatment_allocations, expected_treatment_allocations)

    def test_get_treatment_allocations_auto_all_vs_all(self):
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(3, "all_vs_all", None, None)
        expected_treatment_allocations = np.array((1 / 3, 1 / 3, 1 / 3))
        assert np.array_equal(returned_treatment_allocations, expected_treatment_allocations)

    def test_get_treatment_allocations_auto_control_vs_all_custom_cost(self):
        custom_cost = np.array((1, 2, 2))
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(
            3, "control_vs_all", custom_cost, None
        )
        ratios = np.array((1, 1 / 2 * np.sqrt(1 / 2), 1 / 2 * np.sqrt(1 / 2)))
        expected_treatment_allocations = ratios / np.sum(ratios)
        assert np.array_equal(returned_treatment_allocations, expected_treatment_allocations)

    def test_get_treatment_allocations_auto_all_vs_all_custom_cost(self):
        custom_cost = np.array((1, 2, 2))
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(3, "all_vs_all", custom_cost, None)
        ratios = np.array((1, np.sqrt(1 / 2), np.sqrt(1 / 2)))
        expected_treatment_allocations = ratios / np.sum(ratios)
        assert np.array_equal(returned_treatment_allocations, expected_treatment_allocations)

    def test_get_treatment_allocations_custom(self):
        custom_treatment_allocations = np.array((0.34, 0.56, 0.10))
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(
            3, "control_vs_all", None, custom_treatment_allocations
        )
        assert np.array_equal(returned_treatment_allocations, custom_treatment_allocations)

    def test_get_treatment_allocations_from_custom_list(self):
        custom_treatment_allocations = [0.34, 0.56, 0.10]
        expected_treatment_allocations = np.array(custom_treatment_allocations)
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(
            3, "control_vs_all", None, custom_treatment_allocations
        )
        print("------------- returned:", returned_treatment_allocations)
        assert np.array_equal(returned_treatment_allocations, expected_treatment_allocations)

    def test_get_treatment_allocations_from_custom_tuple(self):
        custom_treatment_allocations = (0.34, 0.56, 0.10)
        expected_treatment_allocations = np.array(custom_treatment_allocations)
        returned_treatment_allocations = conf.SampleSize._get_treatment_allocations(
            3, "control_vs_all", None, custom_treatment_allocations
        )
        print("------------- returned:", returned_treatment_allocations)
        assert np.array_equal(returned_treatment_allocations, expected_treatment_allocations)

    def test_get_treatment_allocations_custom_negative(self):
        with pytest.raises(ValueError):
            custom_treatment_allocations = np.array((0.34, 0.56, -0.10))
            conf.SampleSize._get_treatment_allocations(3, "control_vs_all", None, custom_treatment_allocations)

    def test_get_treatment_allocations_custom_sum_too_large(self):
        with pytest.raises(ValueError):
            custom_treatment_allocations = np.array((0.35, 0.56, 0.10))
            conf.SampleSize._get_treatment_allocations(3, "control_vs_all", None, custom_treatment_allocations)

    def test_get_treatment_allocations_custom_wrong_shape(self):
        with pytest.raises(TypeError):
            custom_treatment_allocations = np.array((0.34, 0.56, 0.05, 0.05))
            conf.SampleSize._get_treatment_allocations(3, "control_vs_all", None, custom_treatment_allocations)

    def test_get_alpha_no_bonferroni(self):
        alpha = conf.SampleSize._get_alpha(0.13, 0.85, False, 5, "control_vs_all")
        assert alpha == 0.13

    def test_get_alpha_bonferroni_control_vs_all(self):
        alpha = conf.SampleSize._get_alpha(0.13, 0.85, True, 5, "control_vs_all")
        assert alpha == 0.13 / 4

    def test_get_alpha_bonferroni_all_vs_all(self):
        alpha = conf.SampleSize._get_alpha(0.13, 0.85, True, 5, "all_vs_all")
        assert alpha == 0.13 / 10

    def test_get_alpha_wrong_bonferroni_type(self):
        with pytest.raises(TypeError):
            conf.SampleSize._get_alpha(0.13, 0.85, 0.0, 5, "control_vs_all")

    def test_get_alpha_out_of_bounds(self):
        with pytest.raises(ValueError):
            conf.SampleSize._get_alpha(1.13, 0.85, False, 5, "control_vs_all")

    def test_get_alpha_type(self):
        with pytest.raises(TypeError):
            conf.SampleSize._get_alpha(np.array((0.13)), 0.85, False, 5, "control_vs_all")

    def test_get_alpha_low_power(self):
        with pytest.raises(ValueError):
            conf.SampleSize._get_alpha(0.13, 0.10, False, 5, "control_vs_all")

    def test_validate_percentage(self):
        num = conf.SampleSize._validate_percentage(0.47)
        assert num == 0.47

    def test_validate_percentage_out_of_bounds(self):
        with pytest.raises(ValueError):
            conf.SampleSize._validate_percentage(1.47)

    def test_validate_percentage_type(self):
        with pytest.raises(TypeError):
            conf.SampleSize._validate_percentage(np.array(0.47))

    def test_validate_positive(self):
        val = 0.3
        return_val = conf.SampleSize._validate_positive(val)
        assert return_val == val

    def test_validate_positive_infinity(self):
        val = np.inf
        return_val = conf.SampleSize._validate_positive(val)
        assert return_val == val

    def test_validate_positive_negative(self):
        with pytest.raises(ValueError):
            conf.SampleSize._validate_positive(-0.3)

    def test_validate_positive_type(self):
        with pytest.raises(TypeError):
            conf.SampleSize._validate_positive("X")

    def test_clean_continuous_mde(self):
        mde = conf.SampleSize._clean_continuous_mde(3.44)
        assert mde == 3.44

    def test_clean_continuous_mde_zero(self):
        with pytest.raises(ValueError):
            conf.SampleSize._clean_continuous_mde(0)

    def test_clean_binomial_mde(self):
        mde = conf.SampleSize._clean_binomial_mde(0.03, 0.47)
        assert mde == 0.03

    def test_clean_binomial_large(self):
        mde = conf.SampleSize._clean_binomial_mde(0.49, 0.47)
        assert mde == 0.49

    def test_clean_binomial_too_large(self):
        with pytest.raises(ValueError):
            conf.SampleSize._clean_binomial_mde(0.54, 0.47)



from typing import Tuple, Dict

import numpy as np
from pandas import DataFrame, Series
from statsmodels.stats.weightstats import _tconfint_generic, _tstat_generic

from spotify_confidence.analysis.confidence_utils import power_calculation
from spotify_confidence.analysis.constants import (
    NUMERATOR,
    NUMERATOR_SUM_OF_SQUARES,
    DENOMINATOR,
    INTERVAL_SIZE,
    POINT_ESTIMATE,
    CI_LOWER,
    CI_UPPER,
    VARIANCE,
    TWO_SIDED,
    SFX1,
    SFX2,
    STD_ERR,
    PREFERENCE_TEST,
    NULL_HYPOTHESIS,
    DIFFERENCE,
)


def point_estimate(df: DataFrame, **kwargs: Dict[str, str]) -> float:
    numerator = kwargs[NUMERATOR]
    denominator = kwargs[DENOMINATOR]
    if (df[denominator] == 0).any():
        raise ValueError("""Can't compute point estimate: denominator is 0""")
    return df[numerator] / df[denominator]


def variance(df: DataFrame, **kwargs: Dict[str, str]) -> float:
    numerator = kwargs[NUMERATOR]
    denominator = kwargs[DENOMINATOR]
    numerator_sumsq = kwargs[NUMERATOR_SUM_OF_SQUARES]
    binary = df[numerator_sumsq] == df[numerator]
    if binary.all():
        # This equals row[POINT_ESTIMATE]*(1-row[POINT_ESTIMATE]) when the data is binary,
        # and also gives a robust fallback in case it's not
        variance = df[numerator_sumsq] / df[denominator] - df[POINT_ESTIMATE] ** 2
    else:
        variance = (df[numerator_sumsq] - np.power(df[numerator], 2) / df[denominator]) / (df[denominator] - 1)
    if (variance < 0).any():
        raise ValueError("Computed variance is negative. " "Please check your inputs.")
    return variance


def std_err(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    denominator = kwargs[DENOMINATOR]
    return np.sqrt(df[VARIANCE + SFX1] / df[denominator + SFX1] + df[VARIANCE + SFX2] / df[denominator + SFX2])


def add_point_estimate_ci(df: DataFrame, **kwargs: Dict[str, str]) -> Series:
    denominator = kwargs[DENOMINATOR]
    interval_size = kwargs[INTERVAL_SIZE]
    df[CI_LOWER], df[CI_UPPER] = _tconfint_generic(
        mean=df[POINT_ESTIMATE],
        std_mean=np.sqrt(df[VARIANCE] / df[denominator]),
        dof=df[denominator] - 1,
        alpha=1 - interval_size,
        alternative=TWO_SIDED,
    )
    return df


def _dof(row: Series, **kwargs: Dict[str, str]) -> float:
    denominator = kwargs[DENOMINATOR]
    v1, v2 = row[VARIANCE + SFX1], row[VARIANCE + SFX2]
    n1, n2 = row[denominator + SFX1], row[denominator + SFX2]
    return (v1 / n1 + v2 / n2) ** 2 / ((v1 / n1) ** 2 / (n1 - 1) + (v2 / n2) ** 2 / (n2 - 1))


def p_value(df: Series, **kwargs: Dict[str, str]) -> Series:
    _, p_value = _tstat_generic(
        value1=df[POINT_ESTIMATE + SFX2],
        value2=df[POINT_ESTIMATE + SFX1],
        std_diff=df[STD_ERR],
        dof=_dof(df, **kwargs),
        alternative=df[PREFERENCE_TEST].values[0],
        diff=df[NULL_HYPOTHESIS],
    )
    return p_value


def ci(df: DataFrame, alpha_column: str, **kwargs: Dict[str, str]) -> Tuple[Series, Series]:
    return _tconfint_generic(
        mean=df[DIFFERENCE],
        std_mean=df[STD_ERR],
        dof=_dof(df, **kwargs),
        alpha=df[alpha_column],
        alternative=df[PREFERENCE_TEST].values[0],
    )


def achieved_power(df: DataFrame, mde: float, alpha: float, **kwargs: Dict[str, str]) -> DataFrame:
    v1, v2 = df[VARIANCE + SFX1], df[VARIANCE + SFX2]
    d1, d2 = kwargs[DENOMINATOR] + SFX1, kwargs[DENOMINATOR] + SFX2
    n1, n2 = df[d1], df[d2]

    var_pooled = ((n1 - 1) * v1 + (n2 - 1) * v2) / (n1 + n2 - 2)

    return power_calculation(mde, var_pooled, alpha, n1, n2)



"""Tests for `confidence` categorical variables."""

import pytest
import spotify_confidence
import pandas as pd
import numpy as np

spotify_confidence.options.set_option("randomization_seed", 1)


class TestCategorical(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "success": [500, 42, 1005, 50, 4, 100],
                "total": [1009, 104, 1502, 100, 10, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.BetaBinomial(
            self.data,
            numerator_column="success",
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_summary(self):
        """Area plot tests"""

        summary = self.test.summary()
        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.success / self.data.total)
        assert np.allclose(
            summary["ci_lower"], np.array([0.46473987, 0.3132419, 0.64500415, 0.40317395, 0.1530671, 0.58861002])
        )
        assert np.allclose(
            summary["ci_upper"], np.array([0.52636595, 0.49971958, 0.69256054, 0.59682605, 0.69632051, 0.73836135])
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1
        ch = self.test.summary_plot(groupby="country")
        assert len(ch.charts) == 2

    def test_difference(self):
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], 0.090737)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.0092189024)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.18712235)
        assert np.allclose(diff["P(level_2 > level_1)"].iloc[0], 0.962782)
        assert np.allclose(diff["level_1 potential loss"].iloc[0], -0.091504838)
        assert np.allclose(diff["level_1 potential gain"].iloc[0], 0.000767784)
        assert np.allclose(diff["level_2 potential loss"].iloc[0], -0.000767784)
        assert np.allclose(diff["level_2 potential gain"].iloc[0], 0.091504838)

        diff = self.test.difference(("us", "control"), ("us", "test"), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], 0.2416812)
        assert np.allclose(diff["P(level_2 > level_1)"].iloc[0], 0.962782)
        assert np.allclose(diff["level_1 potential loss"].iloc[0], -0.2431594)

        diff = self.test.difference("control", "test", groupby="country", absolute=False)
        assert np.allclose(diff["difference"], np.array([0.428745, 0.241799]))
        assert np.allclose(diff["P(level_2 > level_1)"], np.array([0.725982, 0.963228]))
        assert np.allclose(diff["level_2 potential loss"], np.array([-0.042689, -0.001464]), rtol=1e-05, atol=1e-06)

    def test_difference_with_interval_sizes(self):
        self.test._interval_size = 0.99
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], 0.090737)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.040760)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.21614)

        self.test._interval_size = 0.999
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], 0.090737)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.0769830)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.2479583)

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("bad_value", "bad_value"), ("bad_value", "bad_value"))

        ch = self.test.difference_plot(("us", "control"), ("us", "test"))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="country")
        assert len(ch.charts) == 2

    def test_multiple_difference_joint(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_joint(("bad_value", "bad_value"))

        diff = self.test.multiple_difference_joint(("us", "test2"))
        print(np.random.get_state()[1][0])
        assert np.allclose(diff["difference"], np.array([0.000743]), rtol=1e-05, atol=1e-06)
        assert np.allclose(diff["P(('us', 'test2') >= all)"], np.array([0.508644]))
        assert np.allclose(diff["('us', 'test2') potential loss"], np.array([-0.032459]))

        diff = self.test.multiple_difference_joint("test2", groupby="country")
        assert np.allclose(diff["test2 potential loss"], np.array([-0.054338, 0.0]))

    def test_multiple_difference_joint_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_joint_plot(("bad_value", "bad_value"), ("bad_value", "bad_value"))

        ch = self.test.multiple_difference_joint_plot(("us", "control"))
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_joint_plot("test", groupby="country")
        assert len(ch.charts) == 2

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference("bad_value")

        md = self.test.multiple_difference("control", groupby="country")
        for i, row in md.iterrows():
            diff_value = self.test.difference((row["country"], row["level_1"]), (row["country"], row["level_2"]))[
                "difference"
            ].values[0]
            assert np.allclose(row["difference"], diff_value, rtol=1e-02)

    def test_multiple_difference_plot(self):
        ch = self.test.multiple_difference_plot(("us", "control"))
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot("test", groupby="country")
        assert len(ch.charts) == 2

    def test_multiple_difference_level_as_reference(self):
        md = self.test.multiple_difference("control", groupby="country", level_as_reference=True)
        for i, row in md.iterrows():
            diff_value = self.test.difference((row["country"], row["level_1"]), (row["country"], row["level_2"]))[
                "difference"
            ].values[0]
            assert np.allclose(row["difference"], diff_value, rtol=1e-02)


class TestOrdinal:
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "success": [
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                ],
                "total": [
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
            }
        )

        self.test = spotify_confidence.BetaBinomial(
            self.data,
            numerator_column="success",
            denominator_column="total",
            categorical_group_columns="variation_name",
            ordinal_group_column="days_since_reg",
        )

    def test_summary(self):
        """Area plot tests"""
        summary = self.test.summary()
        assert np.array_equal(
            summary["point_estimate"], summary[self.test._numerator_column] / summary[self.test._denominator_column]
        )
        assert np.allclose(
            summary["ci_lower"],
            np.array(
                [
                    0.46426613,
                    0.18932289,
                    0.58861002,
                    0.479025,
                    0.21062941,
                    0.57588269,
                    0.47434292,
                    0.21431068,
                    0.59864266,
                    0.49901672,
                    0.17227621,
                    0.58861002,
                    0.47924165,
                    0.19940202,
                    0.5876837,
                ]
            ),
        )
        assert np.allclose(
            summary["ci_upper"],
            np.array(
                [
                    0.52586121,
                    0.57128953,
                    0.73836135,
                    0.54091721,
                    0.61607968,
                    0.72555745,
                    0.53533858,
                    0.59380455,
                    0.74548633,
                    0.56080991,
                    0.56776609,
                    0.73836135,
                    0.53993569,
                    0.592895,
                    0.73526909,
                ]
            ),
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"))

        diff = self.test.difference(("control", 1), ("test", 1))
        # Add more assertions here

        diff = self.test.difference(("control", 1), ("test", 1), absolute=False)
        # Add more assertions here

        diff = self.test.difference("control", "test", groupby="days_since_reg")
        # Add more assertions here
        assert len(diff) > 0

    def test_multiple_difference(self):
        diff = self.test.multiple_difference(level="control", groupby="days_since_reg", level_as_reference=True)
        assert len(diff) == 10

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot("control", "not_a_valid_level")

        ch = self.test.difference_plot(("control", 1), ("test", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="days_since_reg")
        assert len(ch.charts) == 1

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(level="control")

        ch = self.test.multiple_difference_plot(level="control", groupby=["days_since_reg"], level_as_reference=True)
        assert len(ch.charts) == 1


class TestOrdinalPlusTwoCategorical(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "success": [
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                ],
                "total": [
                    2010,
                    42,
                    250,
                    2000,
                    40,
                    253,
                    2030,
                    43,
                    254,
                    2000,
                    40,
                    250,
                    2040,
                    41,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "days_since_reg": [
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                ],
                "country": [
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.BetaBinomial(
            self.data,
            numerator_column="success",
            denominator_column="total",
            categorical_group_columns=["variation_name", "country"],
            ordinal_group_column="days_since_reg",
        )

    def test_summary_plot(self):
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot(groupby=["country"])
        assert len(ch.charts) == 2

        ch = self.test.summary_plot(groupby=["days_since_reg"])
        assert len(ch.charts) == 5

        ch = self.test.summary_plot(groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 10

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"), groupby="days_since_reg")

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby="days_since_reg")
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby=["days_since_reg"])
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", 1), level_2=("control", 1), groupby=["country"])
        assert len(df) == 2
        assert "country" in df.columns

        df = self.test.difference(level_1="test", level_2="control", groupby=["country", "days_since_reg"])
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", "us", 10), ("test", "us", 10))

        ch = self.test.difference_plot(level_1=("control", "us", 1), level_2=("test", "us", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", 1), level_2=("test", 1), groupby="country")
        assert len(ch.charts) == 2

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(level="control")

        ch = self.test.multiple_difference_plot(
            level="control", groupby=["country", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Iterable, List, Tuple, Dict

import numpy as np
from numpy import isnan
from pandas import DataFrame
from scipy import stats as st

from spotify_confidence.analysis.abstract_base_classes.confidence_computer_abc import ConfidenceComputerABC
from spotify_confidence.analysis.confidence_utils import (
    get_remaning_groups,
    validate_levels,
    level2str,
    listify,
    validate_and_rename_columns,
    drop_and_rename_columns,
    get_all_categorical_group_columns,
    get_all_group_columns,
    validate_data,
    remove_group_columns,
    groupbyApplyParallel,
    reset_named_indices,
)
from spotify_confidence.analysis.constants import (
    NUMERATOR,
    NUMERATOR_SUM_OF_SQUARES,
    DENOMINATOR,
    BOOTSTRAPS,
    INTERVAL_SIZE,
    POINT_ESTIMATE,
    FINAL_EXPECTED_SAMPLE_SIZE,
    ORDINAL_GROUP_COLUMN,
    MDE,
    METHOD,
    CORRECTION_METHOD,
    ABSOLUTE,
    VARIANCE,
    NUMBER_OF_COMPARISONS,
    FEATURE,
    FEATURE_SUMSQ,
    FEATURE_CROSS,
    CI_LOWER,
    CI_UPPER,
    DIFFERENCE,
    P_VALUE,
    SFX1,
    SFX2,
    STD_ERR,
    ADJUSTED_ALPHA_POWER_SAMPLE_SIZE,
    POWER,
    POWERED_EFFECT,
    ADJUSTED_POWER,
    ADJUSTED_P,
    ADJUSTED_LOWER,
    ADJUSTED_UPPER,
    IS_SIGNIFICANT,
    REQUIRED_SAMPLE_SIZE,
    REQUIRED_SAMPLE_SIZE_METRIC,
    NULL_HYPOTHESIS,
    ALTERNATIVE_HYPOTHESIS,
    NIM,
    PREFERENCE,
    PREFERENCE_TEST,
    TWO_SIDED,
    PREFERENCE_DICT,
    CORRECTION_METHODS,
    BOOTSTRAP,
    CHI2,
    TTEST,
    ZTEST,
    NIM_TYPE,
    NIM_COLUMN_DEFAULT,
    PREFERRED_DIRECTION_COLUMN_DEFAULT,
    ZTESTLINREG,
    ORIGINAL_POINT_ESTIMATE,
    ORIGINAL_VARIANCE,
    VARIANCE_REDUCTION,
)
from spotify_confidence.analysis.frequentist.confidence_computers import confidence_computers
from spotify_confidence.analysis.frequentist.multiple_comparison import (
    get_num_comparisons,
    add_adjusted_p_and_is_significant,
    add_ci,
    set_alpha_and_adjust_preference,
    get_preference,
    add_adjusted_power,
)
from spotify_confidence.analysis.frequentist.nims_and_mdes import (
    add_nim_input_columns_from_tuple_or_dict,
    add_nims_and_mdes,
)


class ConfidenceComputer(ConfidenceComputerABC):
    def __init__(
        self,
        data_frame: DataFrame,
        numerator_column: Union[str, None],
        numerator_sum_squares_column: Union[str, None],
        denominator_column: Union[str, None],
        categorical_group_columns: Union[str, Iterable],
        ordinal_group_column: Union[str, None],
        interval_size: float,
        correction_method: str,
        method_column: str,
        bootstrap_samples_column: Union[str, None],
        metric_column: Union[str, None],
        treatment_column: Union[str, None],
        power: float,
        feature_column: Union[str, None],
        feature_sum_squares_column: Union[str, None],
        feature_cross_sum_column: Union[str, None],
    ):
        self._df = data_frame.reset_index(drop=True)
        self._numerator = numerator_column
        self._numerator_sumsq = numerator_sum_squares_column
        if self._numerator is not None and (self._numerator_sumsq is None or self._numerator_sumsq == self._numerator):
            if (data_frame[numerator_column] <= data_frame[denominator_column]).all():
                # Treat as binomial data
                self._numerator_sumsq = self._numerator
            else:
                raise ValueError(
                    f"numerator_sum_squares_column missing or same as "
                    f"numerator_column, but since {numerator_column} is not "
                    f"always smaller than {denominator_column} it can't be "
                    f"binomial data. Please check your data."
                )

        self._denominator = denominator_column
        self._categorical_group_columns = get_all_categorical_group_columns(
            categorical_group_columns, metric_column, treatment_column
        )
        self._segments = remove_group_columns(self._categorical_group_columns, metric_column)
        self._segments = remove_group_columns(self._segments, treatment_column)
        self._ordinal_group_column = ordinal_group_column
        self._metric_column = metric_column
        self._interval_size = interval_size
        self._power = power
        self._treatment_column = treatment_column
        self._feature = feature_column
        self._feature_ssq = feature_sum_squares_column
        self._feature_cross = feature_cross_sum_column

        if correction_method.lower() not in CORRECTION_METHODS:
            raise ValueError(f"Use one of the correction methods in {CORRECTION_METHODS}")
        self._correction_method = correction_method
        self._method_column = method_column

        self._single_metric = False
        if self._metric_column is not None and data_frame.groupby(self._metric_column, sort=False).ngroups == 1:
            self._single_metric = True

        self._all_group_columns = get_all_group_columns(self._categorical_group_columns, self._ordinal_group_column)
        self._bootstrap_samples_column = bootstrap_samples_column

        columns_that_must_exist = []
        if (
            CHI2 in self._df[self._method_column]
            or TTEST in self._df[self._method_column]
            or ZTEST in self._df[self._method_column]
        ):
            columns_that_must_exist += [self._numerator, self._denominator]
            columns_that_must_exist += [] if self._numerator_sumsq is None else [self._numerator_sumsq]
        if BOOTSTRAP in self._df[self._method_column]:
            columns_that_must_exist += [self._bootstrap_samples_column]
        if ZTESTLINREG in self._df[self._method_column]:
            columns_that_must_exist += [self._feature, self._feature_ssq, self._feature_cross]

        validate_data(self._df, columns_that_must_exist, self._all_group_columns, self._ordinal_group_column)

        self._sufficient = None

    def compute_summary(self, verbose: bool) -> DataFrame:
        return (
            self._sufficient_statistics
            if verbose
            else self._sufficient_statistics[
                self._all_group_columns
                + ([self._metric_column] if self._metric_column is not None and self._single_metric else [])
                + [c for c in [self._numerator, self._denominator] if c is not None]
                + [POINT_ESTIMATE, CI_LOWER, CI_UPPER]
            ]
        )

    @property
    def _sufficient_statistics(self) -> DataFrame:
        if self._sufficient is None:
            kwargs = {
                NUMERATOR: self._numerator,
                NUMERATOR_SUM_OF_SQUARES: self._numerator_sumsq,
                DENOMINATOR: self._denominator,
                BOOTSTRAPS: self._bootstrap_samples_column,
                INTERVAL_SIZE: self._interval_size,
                FEATURE: self._feature,
                FEATURE_SUMSQ: self._feature_ssq,
                FEATURE_CROSS: self._feature_cross,
            }
            groupby = [col for col in [self._method_column, self._metric_column] if col is not None]
            self._sufficient = (
                self._df.groupby(groupby, sort=False, group_keys=True)
                .apply(
                    lambda df: df.assign(
                        **{
                            POINT_ESTIMATE: lambda df: confidence_computers[
                                df[self._method_column].values[0]
                            ].point_estimate(df, **kwargs)
                        }
                    )
                    .assign(
                        **{
                            ORIGINAL_POINT_ESTIMATE: lambda df: (
                                confidence_computers[ZTEST].point_estimate(df, **kwargs)
                                if df[self._method_column].values[0] == ZTESTLINREG
                                else confidence_computers[df[self._method_column].values[0]].point_estimate(
                                    df, **kwargs
                                )
                            )
                        }
                    )
                    .assign(
                        **{
                            VARIANCE: lambda df: confidence_computers[df[self._method_column].values[0]].variance(
                                df, **kwargs
                            )
                        }
                    )
                    .assign(
                        **{
                            ORIGINAL_VARIANCE: lambda df: (
                                confidence_computers[ZTEST].variance(df, **kwargs)
                                if df[self._method_column].values[0] == ZTESTLINREG
                                else confidence_computers[df[self._method_column].values[0]].variance(df, **kwargs)
                            )
                        }
                    )
                    .pipe(
                        lambda df: confidence_computers[df[self._method_column].values[0]].add_point_estimate_ci(
                            df, **kwargs
                        )
                    )
                )
                .pipe(reset_named_indices)
            )
        return self._sufficient

    def compute_difference(
        self,
        level_1: Union[str, Iterable],
        level_2: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        difference_df = self._compute_differences(
            level_columns=level_columns,
            levels=[(level_1, level_2)],
            absolute=absolute,
            groupby=groupby,
            level_as_reference=True,
            nims=nims,
            final_expected_sample_size_column=final_expected_sample_size_column,
            mde_column=mde_column,
        )
        return (
            difference_df
            if verbose
            else difference_df[
                listify(groupby)
                + ["level_1", "level_2", "absolute_difference", DIFFERENCE, CI_LOWER, CI_UPPER, P_VALUE]
                + [ADJUSTED_LOWER, ADJUSTED_UPPER, ADJUSTED_P, IS_SIGNIFICANT, POWERED_EFFECT, REQUIRED_SAMPLE_SIZE]
                + ([NIM, NULL_HYPOTHESIS, PREFERENCE] if nims is not None else [])
            ]
        )

    def compute_multiple_difference(
        self,
        level: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        other_levels = [
            other
            for other in self._sufficient_statistics.groupby(level_columns, sort=False).groups.keys()
            if other != level
        ]
        levels = [(level, other) for other in other_levels]
        difference_df = self._compute_differences(
            level_columns=level_columns,
            levels=levels,
            absolute=absolute,
            groupby=groupby,
            level_as_reference=level_as_reference,
            nims=nims,
            final_expected_sample_size_column=final_expected_sample_size_column,
            mde_column=mde_column,
        )
        return (
            difference_df
            if verbose
            else difference_df[
                listify(groupby)
                + [
                    "level_1",
                    "level_2",
                    "absolute_difference",
                    DIFFERENCE,
                    CI_LOWER,
                    CI_UPPER,
                    P_VALUE,
                    POWERED_EFFECT,
                    REQUIRED_SAMPLE_SIZE,
                ]
                + [ADJUSTED_LOWER, ADJUSTED_UPPER, ADJUSTED_P, IS_SIGNIFICANT]
                + ([NIM, NULL_HYPOTHESIS, PREFERENCE] if nims is not None else [])
            ]
        )

    def compute_differences(
        self,
        levels: List[Tuple],
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        verbose: bool,
        mde_column: str,
    ) -> DataFrame:
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        difference_df = self._compute_differences(
            level_columns=level_columns,
            levels=[levels] if type(levels) == tuple else levels,
            absolute=absolute,
            groupby=groupby,
            level_as_reference=True,
            nims=nims,
            final_expected_sample_size_column=final_expected_sample_size_column,
            mde_column=mde_column,
        )
        return (
            difference_df
            if verbose
            else difference_df[
                listify(groupby)
                + ["level_1", "level_2", "absolute_difference", DIFFERENCE, CI_LOWER, CI_UPPER, P_VALUE]
                + [ADJUSTED_LOWER, ADJUSTED_UPPER, ADJUSTED_P, IS_SIGNIFICANT, POWERED_EFFECT, REQUIRED_SAMPLE_SIZE]
                + ([NIM, NULL_HYPOTHESIS, PREFERENCE] if nims is not None else [])
            ]
        )

    def _compute_differences(
        self,
        level_columns: Iterable,
        levels: Union[str, Iterable],
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        nims: NIM_TYPE,
        final_expected_sample_size_column: str,
        mde_column: str,
    ):
        if type(level_as_reference) is not bool:
            raise ValueError(f"level_as_reference must be either True or False, but is {level_as_reference}.")
        groupby = listify(groupby)
        unique_levels = set([level[0] for level in levels] + [level[1] for level in levels])
        validate_levels(self._sufficient_statistics, level_columns, unique_levels)
        str2level = {level2str(lv): lv for lv in unique_levels}
        levels = [
            (level2str(lvl[0]), level2str(lvl[1])) if level_as_reference else (level2str(lvl[1]), level2str(lvl[0]))
            for lvl in levels
        ]

        def assign_total_denominator(df, groupby):
            if self._denominator is None:
                return df.assign(**{f"current_total_{self._denominator}": None})

            if len(groupby) == 0:
                return df.assign(
                    **{f"current_total_{self._denominator}": self._sufficient_statistics[self._denominator].sum()}
                )
            else:
                return df.merge(
                    df.groupby(groupby, sort=False)[self._denominator]
                    .sum()
                    .reset_index()
                    .rename(columns={self._denominator: f"current_total_{self._denominator}"})
                )

        return (
            self._sufficient_statistics.assign(
                level=self._sufficient_statistics[level_columns].agg(level2str, axis="columns")
            )
            .pipe(assign_total_denominator, groupby)
            .query(f"level in {[l1 for l1,l2 in levels] + [l2 for l1,l2 in levels]}")
            .pipe(lambda df: df if groupby == [] else df.set_index(groupby))
            .pipe(
                self._create_comparison_df,
                groups_to_compare=levels,
                absolute=absolute,
                nims=nims,
                mde_column=mde_column,
                final_expected_sample_size_column=final_expected_sample_size_column,
            )
            .assign(level_1=lambda df: df["level_1"].map(lambda s: str2level[s]))
            .assign(level_2=lambda df: df["level_2"].map(lambda s: str2level[s]))
            .pipe(lambda df: df.reset_index([name for name in df.index.names if name is not None]))
            .reset_index(drop=True)
            .sort_values(by=groupby + ["level_1", "level_2"])
            .reset_index(drop=True)
        )

    def _create_comparison_df(
        self,
        df: DataFrame,
        groups_to_compare: List[Tuple[str, str]],
        absolute: bool,
        nims: NIM_TYPE,
        mde_column: bool,
        final_expected_sample_size_column: str,
    ) -> DataFrame:
        def join(df: DataFrame) -> DataFrame:
            has_index = not all(idx is None for idx in df.index.names)
            if has_index:
                # self-join on index (the index will typically model the date,
                # i.e., rows with the same date are joined)
                return df.merge(df, left_index=True, right_index=True, suffixes=(SFX1, SFX2))
            else:
                # join on dummy column, i.e. conduct a cross join
                return (
                    df.assign(dummy_join_column=1)
                    .merge(right=df.assign(dummy_join_column=1), on="dummy_join_column", suffixes=(SFX1, SFX2))
                    .drop(columns="dummy_join_column")
                )

        comparison_df = (
            df.pipe(add_nim_input_columns_from_tuple_or_dict, nims=nims, mde_column=mde_column)
            .pipe(
                add_nims_and_mdes,
                mde_column=mde_column,
                nim_column=NIM_COLUMN_DEFAULT,
                preferred_direction_column=PREFERRED_DIRECTION_COLUMN_DEFAULT,
            )
            .pipe(join)
            .query(
                "("
                + " or ".join([f"(level_1=='{l1}' and level_2=='{l2}')" for l1, l2 in groups_to_compare])
                + ")"
                + "and level_1 != level_2"
            )
            .pipe(
                validate_and_rename_columns,
                [NIM, mde_column, PREFERENCE, final_expected_sample_size_column, self._method_column],
            )
            .pipe(
                drop_and_rename_columns,
                [NULL_HYPOTHESIS, ALTERNATIVE_HYPOTHESIS, f"current_total_{self._denominator}"],
            )
            .assign(**{PREFERENCE_TEST: lambda df: get_preference(df, self._correction_method)})
            .assign(**{POWER: self._power})
            .pipe(
                add_adjusted_power,
                correction_method=self._correction_method,
                metric_column=self._metric_column,
                single_metric=self._single_metric,
            )
        )

        groups_except_ordinal = [
            column
            for column in df.index.names
            if column is not None
            and (column != self._ordinal_group_column or final_expected_sample_size_column is None)
        ]
        n_comparisons = get_num_comparisons(
            comparison_df,
            self._correction_method,
            number_of_level_comparisons=comparison_df.groupby(["level_1", "level_2"], sort=False).ngroups,
            groupby=groups_except_ordinal,
            metric_column=self._metric_column,
            treatment_column=self._treatment_column,
            single_metric=self._single_metric,
            segments=self._segments,
        )

        kwargs = {
            NUMERATOR: self._numerator,
            NUMERATOR_SUM_OF_SQUARES: self._numerator_sumsq,
            DENOMINATOR: self._denominator,
            BOOTSTRAPS: self._bootstrap_samples_column,
            FINAL_EXPECTED_SAMPLE_SIZE: final_expected_sample_size_column,
            ORDINAL_GROUP_COLUMN: self._ordinal_group_column,
            MDE: mde_column,
            METHOD: self._method_column,
            CORRECTION_METHOD: self._correction_method,
            INTERVAL_SIZE: self._interval_size,
            ABSOLUTE: absolute,
            NUMBER_OF_COMPARISONS: n_comparisons,
        }
        comparison_df = groupbyApplyParallel(
            comparison_df.groupby(
                groups_except_ordinal + [self._method_column, "level_1", "level_2"], as_index=False, sort=False
            ),
            lambda df: _compute_comparisons(df, **kwargs),
        )
        return comparison_df

    def achieved_power(self, level_1, level_2, mde, alpha, groupby):
        groupby = listify(groupby)
        level_columns = get_remaning_groups(self._all_group_columns, groupby)
        kwargs = {NUMERATOR: self._numerator, DENOMINATOR: self._denominator}
        return (
            self._compute_differences(
                level_columns,
                [(level_1, level_2)],
                True,
                groupby,
                level_as_reference=True,
                nims=None,
                final_expected_sample_size_column=None,
                mde_column=None,
            )  # TODO: IS this right?
            .pipe(lambda df: df if groupby == [] else df.set_index(groupby))
            .assign(
                achieved_power=lambda df: df.apply(
                    lambda row: confidence_computers[row[self._method_column]].achieved_power(
                        row, mde=mde, alpha=alpha, **kwargs
                    ),
                    axis=1,
                )
            )
        )[["level_1", "level_2", "achieved_power"]]


def _compute_comparisons(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return (
        df.assign(**{DIFFERENCE: lambda df: df[POINT_ESTIMATE + SFX2] - df[POINT_ESTIMATE + SFX1]})
        .assign(**{STD_ERR: confidence_computers[df[kwargs[METHOD]].values[0]].std_err(df, **kwargs)})
        .pipe(_add_p_value_and_ci, **kwargs)
        .pipe(_powered_effect_and_required_sample_size_from_difference_df, **kwargs)
        .pipe(_adjust_if_absolute, absolute=kwargs[ABSOLUTE])
        .assign(**{PREFERENCE: lambda df: df[PREFERENCE].map(PREFERENCE_DICT)})
        .pipe(_add_variance_reduction_rate, **kwargs)
    )


def _add_variance_reduction_rate(df: DataFrame, **kwargs: Dict) -> DataFrame:
    denominator = kwargs[DENOMINATOR]
    method_column = kwargs[METHOD]
    if (df[method_column] == ZTESTLINREG).any():
        variance_no_reduction = (
            df[ORIGINAL_VARIANCE + SFX1] / df[denominator + SFX1]
            + df[ORIGINAL_VARIANCE + SFX2] / df[denominator + SFX2]
        )
        variance_w_reduction = (
            df[VARIANCE + SFX1] / df[denominator + SFX1] + df[VARIANCE + SFX2] / df[denominator + SFX2]
        )
        df = df.assign(**{VARIANCE_REDUCTION: 1 - np.divide(variance_w_reduction, variance_no_reduction)})
    return df


def _add_p_value_and_ci(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return (
        df.pipe(set_alpha_and_adjust_preference, **kwargs)
        .assign(**{P_VALUE: lambda df: df.pipe(_p_value, **kwargs)})
        .pipe(add_adjusted_p_and_is_significant, **kwargs)
        .pipe(add_ci, **kwargs)
    )


def _adjust_if_absolute(df: DataFrame, absolute: bool) -> DataFrame:
    if absolute:
        return df.assign(absolute_difference=absolute)
    else:
        return (
            df.assign(absolute_difference=absolute)
            .assign(**{DIFFERENCE: df[DIFFERENCE] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{CI_LOWER: df[CI_LOWER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{CI_UPPER: df[CI_UPPER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{ADJUSTED_LOWER: df[ADJUSTED_LOWER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{ADJUSTED_UPPER: df[ADJUSTED_UPPER] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{NULL_HYPOTHESIS: df[NULL_HYPOTHESIS] / df[POINT_ESTIMATE + SFX1]})
            .assign(**{POWERED_EFFECT: df[POWERED_EFFECT] / df[POINT_ESTIMATE + SFX1]})
        )


def _p_value(df: DataFrame, **kwargs: Dict) -> float:
    if df[kwargs[METHOD]].values[0] == CHI2 and (df[NIM].notna()).any():
        raise ValueError("Non-inferiority margins not supported in ChiSquared. Use StudentsTTest or ZTest instead.")
    return confidence_computers[df[kwargs[METHOD]].values[0]].p_value(df, **kwargs)


def _powered_effect_and_required_sample_size_from_difference_df(df: DataFrame, **kwargs: Dict) -> DataFrame:
    if df[kwargs[METHOD]].values[0] not in [ZTEST, ZTESTLINREG] and kwargs[MDE] in df:
        raise ValueError("Minimum detectable effects only supported for ZTest.")
    elif df[kwargs[METHOD]].values[0] not in [ZTEST, ZTESTLINREG] or (df[ADJUSTED_POWER].isna()).any():
        df[POWERED_EFFECT] = None
        df[REQUIRED_SAMPLE_SIZE] = None
        df[REQUIRED_SAMPLE_SIZE_METRIC] = None
        return df
    else:
        n1, n2 = df[kwargs[DENOMINATOR] + SFX1], df[kwargs[DENOMINATOR] + SFX2]
        kappa = n1 / n2
        binary = (df[kwargs[NUMERATOR_SUM_OF_SQUARES] + SFX1] == df[kwargs[NUMERATOR] + SFX1]).all()
        proportion_of_total = (n1 + n2) / df[f"current_total_{kwargs[DENOMINATOR]}"]

        z_alpha = st.norm.ppf(
            1
            - df[ADJUSTED_ALPHA_POWER_SAMPLE_SIZE].values[0] / (2 if df[PREFERENCE_TEST].values[0] == TWO_SIDED else 1)
        )
        z_power = st.norm.ppf(df[ADJUSTED_POWER].values[0])

        nim = df[NIM].values[0]
        if isinstance(nim, float):
            non_inferiority = not isnan(nim)
        elif nim is None:
            non_inferiority = nim is not None

        df[POWERED_EFFECT] = confidence_computers[df[kwargs[METHOD]].values[0]].powered_effect(
            df=df.assign(kappa=kappa)
            .assign(current_number_of_units=df[f"current_total_{kwargs[DENOMINATOR]}"])
            .assign(proportion_of_total=proportion_of_total),
            z_alpha=z_alpha,
            z_power=z_power,
            binary=binary,
            non_inferiority=non_inferiority,
            avg_column=ORIGINAL_POINT_ESTIMATE + SFX1,
            var_column=VARIANCE + SFX1,
        )

        if ALTERNATIVE_HYPOTHESIS in df and NULL_HYPOTHESIS in df and (df[ALTERNATIVE_HYPOTHESIS].notna()).all():
            df[REQUIRED_SAMPLE_SIZE] = confidence_computers[df[kwargs[METHOD]].values[0]].required_sample_size(
                proportion_of_total=1,
                z_alpha=z_alpha,
                z_power=z_power,
                binary=binary,
                non_inferiority=non_inferiority,
                hypothetical_effect=df[ALTERNATIVE_HYPOTHESIS] - df[NULL_HYPOTHESIS],
                control_avg=df[ORIGINAL_POINT_ESTIMATE + SFX1],
                control_var=df[VARIANCE + SFX1],
                kappa=kappa,
            )
            df[REQUIRED_SAMPLE_SIZE_METRIC] = confidence_computers[df[kwargs[METHOD]].values[0]].required_sample_size(
                proportion_of_total=proportion_of_total,
                z_alpha=z_alpha,
                z_power=z_power,
                binary=binary,
                non_inferiority=non_inferiority,
                hypothetical_effect=df[ALTERNATIVE_HYPOTHESIS] - df[NULL_HYPOTHESIS],
                control_avg=df[ORIGINAL_POINT_ESTIMATE + SFX1],
                control_var=df[VARIANCE + SFX1],
                kappa=kappa,
            )
        else:
            df[REQUIRED_SAMPLE_SIZE] = None
            df[REQUIRED_SAMPLE_SIZE_METRIC] = None

        return df
